# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import logging
import os
from tempfile import TemporaryDirectory
from time import sleep

import pytest
from test_job import TestJobBase

import signac
from signac import JSONDict, sync
from signac._utility import _mkdir_p
from signac.errors import DocumentSyncConflict, FileSyncConflict, SchemaSyncConflict
from signac.sync import _DocProxy, _FileModifyProxy


def touch(fname, mode=0o666, dir_fd=None, **kwargs):
    """Utility function for updating a file time stamp.

    Source:
        https://stackoverflow.com/questions/1158076/implement-touch-using-python
    """
    flags = os.O_CREAT | os.O_APPEND
    with os.fdopen(os.open(fname, flags=flags, mode=mode, dir_fd=dir_fd)) as f:
        os.utime(
            f.fileno() if os.utime in os.supports_fd else fname,
            dir_fd=None if os.supports_fd else dir_fd,
            **kwargs,
        )


class TestDocProxy:
    def test_basic(self):
        doc = dict(a=0)
        proxy = _DocProxy(doc)
        assert proxy == proxy
        assert proxy == doc
        str(proxy)
        repr(proxy)
        assert len(proxy) == len(doc)
        assert proxy["a"] == doc["a"]
        assert "a" in proxy
        for key in proxy:
            assert key == "a"
        for key in proxy.keys():
            assert key == "a"
        proxy["a"] = 1
        assert proxy["a"] == doc["a"]
        assert proxy == proxy
        assert proxy == doc
        doc["a"] = 2
        proxy.update(doc)
        assert proxy["a"] == doc["a"]
        assert proxy == proxy
        assert proxy == doc

    def test_dry_run(self):
        doc = dict(a=0)
        proxy = _DocProxy(doc, dry_run=True)
        assert proxy == proxy
        assert proxy == doc
        str(proxy)
        repr(proxy)
        assert len(proxy) == len(doc)
        assert proxy["a"] == doc["a"]
        assert "a" in proxy
        for key in proxy:
            assert key == "a"
        for key in proxy.keys():
            assert key == "a"
        proxy["a"] = 1
        assert proxy["a"] == 0
        assert proxy == proxy
        assert proxy == doc


class TestFileModifyProxy:
    def test_copy(self):
        proxy = _FileModifyProxy()
        with TemporaryDirectory(prefix="signac_") as tmp:
            fn_src = os.path.join(tmp, "src.txt")
            fn_dst = os.path.join(tmp, "dst.txt")
            touch(fn_src)
            assert os.path.isfile(fn_src)
            assert not os.path.isfile(fn_dst)
            proxy.copy(fn_src, fn_dst)
            assert os.path.isfile(fn_src)
            assert os.path.isfile(fn_dst)

    def test_copy_dry_run(self):
        proxy = _FileModifyProxy(dry_run=True)
        with TemporaryDirectory(prefix="signac_") as tmp:
            fn_src = os.path.join(tmp, "src.txt")
            fn_dst = os.path.join(tmp, "dst.txt")
            with open(fn_src, "w") as file:
                file.write("test")
            assert os.path.isfile(fn_src)
            assert not os.path.isfile(fn_dst)
            proxy.copy(fn_src, fn_dst)
            assert os.path.isfile(fn_src)
            assert not os.path.isfile(fn_dst)

    def test_copytree(self):
        proxy = _FileModifyProxy()
        with TemporaryDirectory(prefix="signac_") as tmp:
            src = os.path.join(tmp, "src")
            dst = os.path.join(tmp, "dst")
            _mkdir_p(src)
            fn_src = os.path.join(src, "test.txt")
            fn_dst = os.path.join(dst, "test.txt")
            touch(fn_src)
            assert os.path.isfile(fn_src)
            assert not os.path.isfile(fn_dst)
            proxy.copytree(src, dst)
            assert os.path.isfile(fn_src)
            assert os.path.isfile(fn_dst)

    def test_copytree_dryrun(self):
        proxy = _FileModifyProxy(dry_run=True)
        with TemporaryDirectory(prefix="signac_") as tmp:
            src = os.path.join(tmp, "src")
            dst = os.path.join(tmp, "dst")
            _mkdir_p(src)
            fn_src = os.path.join(src, "test.txt")
            fn_dst = os.path.join(dst, "test.txt")
            touch(fn_src)
            assert os.path.isfile(fn_src)
            assert not os.path.isfile(fn_dst)
            proxy.copytree(src, dst)
            assert os.path.isfile(fn_src)
            assert not os.path.isfile(fn_dst)

    def test_remove(self):
        proxy = _FileModifyProxy()
        with TemporaryDirectory(prefix="signac_") as tmp:
            fn = os.path.join(tmp, "test.txt")
            assert not os.path.isfile(fn)
            touch(fn)
            assert os.path.isfile(fn)
            proxy.remove(fn)
            assert not os.path.isfile(fn)

    def test_remove_dryrun(self):
        proxy = _FileModifyProxy(dry_run=True)
        with TemporaryDirectory(prefix="signac_") as tmp:
            fn = os.path.join(tmp, "test.txt")
            assert not os.path.isfile(fn)
            touch(fn)
            assert os.path.isfile(fn)
            proxy.remove(fn)
            assert os.path.isfile(fn)

    def test_create_backup(self):
        proxy = _FileModifyProxy()
        with TemporaryDirectory(prefix="signac_") as tmp:
            fn = os.path.join(tmp, "test.txt")
            assert not os.path.isfile(fn)
            with open(fn, "w") as file:
                file.write("a")
            assert os.path.isfile(fn)
            with proxy.create_backup(fn) as fn_backup:
                assert os.path.isfile(fn_backup)
            assert os.path.isfile(fn)
            assert not os.path.isfile(fn_backup)
            with pytest.raises(RuntimeError):
                with proxy.create_backup(fn) as fn_backup:
                    assert os.path.isfile(fn_backup)
                    with open(fn, "w") as file:
                        file.write("b")
                    raise RuntimeError()
            assert os.path.isfile(fn)
            assert not os.path.isfile(fn_backup)
            with open(fn) as file:
                assert file.read() == "a"

    def test_create_backup_dryrun(self):
        proxy = _FileModifyProxy(dry_run=True)
        with TemporaryDirectory(prefix="signac_") as tmp:
            fn = os.path.join(tmp, "test.txt")
            assert not os.path.isfile(fn)
            with open(fn, "w") as file:
                file.write("a")
            assert os.path.isfile(fn)
            with proxy.create_backup(fn) as fn_backup:
                assert not os.path.isfile(fn_backup)
            assert os.path.isfile(fn)
            assert not os.path.isfile(fn_backup)
            with pytest.raises(RuntimeError):
                with proxy.create_backup(fn) as fn_backup:
                    assert not os.path.isfile(fn_backup)
                    with open(fn, "w") as file:
                        file.write("b")
                    raise RuntimeError()
            assert os.path.isfile(fn)
            assert not os.path.isfile(fn_backup)
            with open(fn) as file:
                assert file.read() == "b"


class TestFileModifyProxyDocBackup:
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.doc = {}

    def test_create_doc_dict(self):
        proxy = _FileModifyProxy()
        with proxy.create_doc_backup(self.doc) as p:
            pass
        with proxy.create_doc_backup(self.doc) as p:
            p["a"] = 0
        assert len(self.doc) == 1
        assert self.doc["a"] == 0

    def test_create_doc_dict_dryrun(self):
        proxy = _FileModifyProxy(dry_run=True)
        with proxy.create_doc_backup(self.doc) as p:
            pass
        with proxy.create_doc_backup(self.doc) as p:
            p["a"] = 0
        assert len(self.doc) == 0

    def test_create_doc_dict_with_error(self):
        proxy = _FileModifyProxy()
        with pytest.raises(RuntimeError):
            with proxy.create_doc_backup(self.doc) as p:
                p["a"] = 0
                raise RuntimeError()
        assert len(self.doc) == 0

    def test_create_doc_dict_with_error_dryrun(self):
        proxy = _FileModifyProxy(dry_run=True)
        with pytest.raises(RuntimeError):
            with proxy.create_doc_backup(self.doc) as p:
                p["a"] = 0
                raise RuntimeError()
        assert len(self.doc) == 0


class TestFileModifyProxyJSONDocBackup(TestFileModifyProxyDocBackup):
    @pytest.fixture(autouse=True)
    def setUp(self, request):
        self._tmp_dir = TemporaryDirectory(prefix="signac_")
        request.addfinalizer(self._tmp_dir.cleanup)
        self.doc = JSONDict(filename=os.path.join(self._tmp_dir.name, "doc.json"))


class TestJobSync(TestJobBase):
    def test_sync_no_implicit_init(self):
        job_dst = self.open_job({"a": 0})
        job_src = self.open_job({"a": 1})
        assert job_dst not in self.project
        assert job_src not in self.project
        job_dst.sync(job_src)
        assert job_dst not in self.project
        assert job_src not in self.project
        job_src.init()
        assert job_src in self.project
        job_dst.sync(job_src)
        assert job_dst in self.project

    def test_file_sync(self):
        job_dst = self.open_job({"a": 0})
        job_src = self.open_job({"a": 1})
        with job_src:
            with open("test", "w") as file:
                file.write("test")
            os.makedirs("subdir")
            with open("subdir/test2", "w") as file:
                file.write("test2")
        assert job_src.isfile("test")
        try:
            logging.disable(logging.WARNING)
            job_dst.sync(job_src)
        finally:
            logging.disable(logging.NOTSET)
        assert job_dst in self.project
        assert job_dst.isfile("test")
        assert not job_dst.isfile("subdir/test2")
        with open(job_dst.fn("test")) as file:
            assert file.read() == "test"

    def test_file_sync_recursive(self):
        job_dst = self.open_job({"a": 0})
        job_src = self.open_job({"a": 1})
        with job_src:
            with open("test", "w") as file:
                file.write("test")
            os.makedirs("subdir")
            with open("subdir/test2", "w") as file:
                file.write("test2")
        assert job_src.isfile("test")
        job_dst.sync(job_src, recursive=True)
        assert job_dst in self.project
        assert job_dst.isfile("test")
        assert job_dst.isfile("subdir/test2")
        with open(job_dst.fn("test")) as file:
            assert file.read() == "test"
        with open(job_dst.fn("subdir/test2")) as file:
            assert file.read() == "test2"

    def test_file_sync_deep(self):
        job_dst = self.open_job({"a": 0})
        job_src = self.open_job({"a": 1})
        with job_src:
            with open("test", "w") as file:
                file.write("test")
            os.makedirs("subdir")
            with open("subdir/test2", "w") as file:
                file.write("test2")
        assert job_src.isfile("test")
        job_dst.sync(job_src, deep=True, recursive=True)
        assert job_dst in self.project
        assert job_dst.isfile("test")
        assert job_dst.isfile("subdir/test2")
        with open(job_dst.fn("test")) as file:
            assert file.read() == "test"
        with open(job_dst.fn("subdir/test2")) as file:
            assert file.read() == "test2"

    def _reset_differing_jobs(self, jobs):
        for i, job in enumerate(jobs):
            with job:
                with open("test", "w") as file:
                    file.write("x" * i)
                _mkdir_p("subdir")
                with open("subdir/test2", "w") as file:
                    file.write("x" * i)

        def differs(fn):
            x = set()
            for job in jobs:
                with open(job.fn(fn)) as file:
                    x.add(file.read())
            return len(x) > 1

        return differs

    def test_file_sync_with_conflict(self):
        job_dst = self.open_job({"a": 0})
        job_src = self.open_job({"a": 1})
        differs = self._reset_differing_jobs((job_dst, job_src))
        assert differs("test")
        assert differs("subdir/test2")
        with pytest.raises(FileSyncConflict):
            job_dst.sync(job_src, recursive=True)
        job_dst.sync(job_src, sync.FileSync.never, recursive=True)
        assert differs("test")
        assert differs("subdir/test2")
        job_dst.sync(job_src, sync.FileSync.always, exclude="test", recursive=True)
        assert differs("test")
        assert differs("subdir/test2")
        job_dst.sync(
            job_src,
            sync.FileSync.always,
            exclude=["test", "non-existent-key"],
            recursive=True,
        )
        assert differs("test")
        assert differs("subdir/test2")
        sleep(1)
        touch(job_src.fn("test"))
        job_dst.sync(job_src, sync.FileSync.update, recursive=True)
        assert not differs("test")
        touch(job_src.fn("subdir/test2"))
        job_dst.sync(job_src, sync.FileSync.update, exclude="test2", recursive=True)
        assert not differs("test")
        job_dst.sync(job_src, sync.FileSync.update, recursive=True)
        assert not differs("subdir/test2")

    def test_file_sync_strategies(self):
        job_dst = self.open_job({"a": 0})
        job_src = self.open_job({"a": 1})

        def reset():
            return self._reset_differing_jobs((job_dst, job_src))

        differs = reset()
        assert differs("test")
        assert differs("subdir/test2")
        with pytest.raises(FileSyncConflict):
            job_dst.sync(job_src, recursive=True)
        assert differs("test")
        assert differs("subdir/test2")
        job_dst.sync(job_src, sync.FileSync.never, recursive=True)
        assert differs("test")
        assert differs("subdir/test2")
        job_dst.sync(job_src, sync.FileSync.always, recursive=True)
        assert not differs("test")
        assert not differs("subdir/test2")
        reset()
        assert differs("test")
        assert differs("subdir/test2")
        sleep(1)
        touch(job_src.fn("test"))
        job_dst.sync(job_src, sync.FileSync.update, recursive=True)
        assert not differs("test")
        touch(job_src.fn("subdir/test2"))
        job_dst.sync(job_src, sync.FileSync.update, recursive=True)
        assert not differs("test")
        assert not differs("subdir/test2")

    def _reset_document_sync(self):
        job_src = self.open_job({"a": 0})
        job_dst = self.open_job({"a": 1})
        job_src.document["a"] = 0
        job_src.document["nested"] = dict(a=1)
        assert job_src.document != job_dst.document
        return job_dst, job_src

    def test_document_sync(self):
        job_dst, job_src = self._reset_document_sync()
        job_dst.sync(job_src)
        assert len(job_dst.document) == len(job_src.document)
        assert job_src.document == job_dst.document
        assert job_src.document["a"] == job_dst.document["a"]
        assert job_src.document["nested"]["a"] == job_dst.document["nested"]["a"]
        job_dst.sync(job_src)
        assert job_src.document == job_dst.document

    def test_document_sync_nested(self):
        job_dst, job_src = self._reset_document_sync()
        job_dst.document["nested"] = dict(a=0)
        with pytest.raises(DocumentSyncConflict):
            job_dst.sync(job_src)
        assert job_src.document != job_dst.document

    def test_document_sync_explicit_overwrit(self):
        job_dst, job_src = self._reset_document_sync()
        job_dst.sync(job_src, doc_sync=sync.DocSync.update)
        assert job_src.document == job_dst.document

    def test_document_sync_overwrite_specific(self):
        job_dst, job_src = self._reset_document_sync()
        job_dst.sync(job_src, doc_sync=sync.DocSync.ByKey("nested.a"))
        assert job_src.document == job_dst.document

    def test_document_sync_partially_differing(self):
        job_dst, job_src = self._reset_document_sync()
        job_dst.document["a"] = 0
        job_dst.sync(job_src)
        assert job_src.document == job_dst.document

    def test_document_sync_differing_keys(self):
        job_dst, job_src = self._reset_document_sync()
        job_src.document["b"] = 1
        job_src.document["nested"]["b"] = 1
        job_dst.sync(job_src)
        assert job_src.document == job_dst.document

    def test_document_sync_no_sync(self):
        job_dst, job_src = self._reset_document_sync()
        assert sync.DocSync.NO_SYNC is False
        job_dst.sync(job_src, doc_sync=False)
        assert job_src.document != job_dst.document
        assert len(job_dst.document) == 0

    def test_document_sync_dst_has_extra_key(self):
        job_dst, job_src = self._reset_document_sync()
        job_dst.document["b"] = 2
        assert "b" not in job_src.document
        assert "b" in job_dst.document
        job_dst.sync(job_src)
        assert "b" not in job_src.document
        assert "b" in job_dst.document
        assert job_dst.document != job_src.document
        assert job_dst.document["nested"] == job_src.document["nested"]
        assert job_dst.document["a"] == job_src.document["a"]

    def test_document_sync_with_error(self):
        job_dst = self.open_job({"a": 0})
        job_src = self.open_job({"a": 1})
        job_dst.document["a"] = 0
        job_src.document["a"] = 1

        def raise_error(src, dst):
            raise RuntimeError()

        with pytest.raises(RuntimeError):
            job_dst.sync(job_src, doc_sync=raise_error)

    def test_document_sync_with_conflict(self):
        job_dst = self.open_job({"a": 0})
        job_src = self.open_job({"a": 1})

        def reset():
            job_src.document["a"] = 0
            job_src.document["nested"] = dict(a=1)
            job_dst.document["a"] = 1
            job_dst.document["nested"] = dict(a=2)

        reset()
        assert job_dst.document != job_src.document
        with pytest.raises(DocumentSyncConflict):
            job_dst.sync(job_src)
        assert job_dst.document != job_src.document
        job_dst.sync(job_src, doc_sync=sync.DocSync.NO_SYNC)
        assert job_dst.document != job_src.document
        assert job_dst.document["a"] != job_src.document["a"]
        assert job_dst.document["nested"] != job_src.document["nested"]
        reset()  # only sync a
        job_dst.sync(job_src, doc_sync=sync.DocSync.ByKey("a"))
        assert job_dst.document != job_src.document
        assert job_dst.document["nested"] != job_src.document["nested"]
        assert job_dst.document["a"] == job_src.document["a"]
        reset()  # only sync nested
        job_dst.sync(job_src, doc_sync=sync.DocSync.ByKey("nested"))
        assert job_dst.document != job_src.document
        assert job_dst.document["a"] != job_src.document["a"]
        assert job_dst.document["nested"] == job_src.document["nested"]
        reset()
        job_dst.sync(job_src, doc_sync=sync.DocSync.ByKey(r"(nested\.)?a"))
        assert job_dst.document == job_src.document
        assert job_dst.document["nested"] == job_src.document["nested"]
        assert job_dst.document["a"] == job_src.document["a"]
        reset()
        job_dst.sync(
            job_src, doc_sync=sync.DocSync.ByKey(lambda key: key.startswith("a"))
        )
        assert job_dst.document != job_src.document
        assert job_dst.document["nested"] != job_src.document["nested"]
        assert job_dst.document["a"] == job_src.document["a"]
        reset()
        job_dst.sync(
            job_src, doc_sync=sync.DocSync.ByKey(lambda key: key.startswith("nested"))
        )
        assert job_dst.document != job_src.document
        assert job_dst.document["a"] != job_src.document["a"]
        assert job_dst.document["nested"] == job_src.document["nested"]
        reset()
        job_dst.sync(job_src, doc_sync=sync.DocSync.update)
        assert job_dst.document == job_src.document
        assert job_dst.document["nested"] == job_src.document["nested"]
        assert job_dst.document["a"] == job_src.document["a"]


class TestProjectSync:
    @pytest.fixture(autouse=True)
    def setUp(self, request):
        self._tmp_dir = TemporaryDirectory(prefix="signac_")
        request.addfinalizer(self._tmp_dir.cleanup)
        self._tmp_pr_a = os.path.join(self._tmp_dir.name, "pr_a")
        self._tmp_pr_b = os.path.join(self._tmp_dir.name, "pr_b")
        os.mkdir(self._tmp_pr_a)
        os.mkdir(self._tmp_pr_b)
        self.project_a = signac.Project.init_project(path=self._tmp_pr_a)
        self.project_b = signac.Project.init_project(path=self._tmp_pr_b)

    def _init_job(self, job, data="data"):
        with job:
            with open("test.txt", "w") as file:
                file.write(str(data))

    def test_src_and_dst_identical(self):
        with pytest.raises(ValueError):
            self.project_a.sync(self.project_a)

    def test_src_and_dst_empty(self):
        self.project_a.sync(self.project_b)
        assert len(self.project_a) == len(self.project_b)

    def test_src_empty(self):
        for i in range(4):
            self._init_job(self.project_b.open_job({"a": i}))
        self.project_a.sync(self.project_b)
        assert len(self.project_a) == len(self.project_b)

    def test_dst_empty(self):
        for i in range(4):
            self._init_job(self.project_a.open_job({"a": i}))
        self.project_a.sync(self.project_b)
        assert len(self.project_a) == 4
        assert len(self.project_b) == 0

    def test_doc_sync(self):
        self.project_a.document["a"] = 0
        assert "a" in self.project_a.document
        assert "a" not in self.project_b.document
        self.project_a.sync(self.project_b)
        assert "a" in self.project_a.document
        assert "a" not in self.project_b.document
        self.project_b.document["b"] = 1
        self.project_a.sync(self.project_b)
        assert "b" in self.project_a.document
        self.project_a.document["b"] = 2
        with pytest.raises(DocumentSyncConflict):
            self.project_a.sync(self.project_b)
        self.project_a.sync(self.project_b, doc_sync=sync.DocSync.ByKey("b"))

    def _setup_mixed(self):
        for i in range(4):
            if i % 2 == 0:
                self._init_job(self.project_a.open_job({"a": i}))
            if i % 3 == 0:
                self._init_job(self.project_b.open_job({"a": i}))

    def test_mixed(self):
        self._setup_mixed()
        with pytest.raises(SchemaSyncConflict):
            self.project_a.sync(self.project_b)
        assert len(self.project_a) == 2
        assert len(self.project_b) == 2
        self.project_a.sync(self.project_b, check_schema=False)
        assert len(self.project_a) == 3

    def _setup_jobs(self):
        for i in range(4):
            self._init_job(self.project_a.open_job({"a": i}))
            self._init_job(self.project_b.open_job({"a": i}))

    def test_with_conflict(self):
        self._setup_jobs()
        assert len(self.project_a) == len(self.project_b)
        job_a0 = self.project_a.open_job({"a": 0})
        with open(job_a0.fn("test.txt"), "w") as file:
            file.write("newdata")
        with pytest.raises(FileSyncConflict):
            self.project_a.sync(self.project_b)

    def test_with_conflict_never(self):
        self._setup_jobs()
        job_a0 = self.project_a.open_job({"a": 0})
        with open(job_a0.fn("text.txt"), "w") as file:
            file.write("otherdata")
        self.project_a.sync(self.project_b, sync.FileSync.never)
        with open(job_a0.fn("text.txt")) as file:
            assert file.read() == "otherdata"

    def test_selection(self):
        self._setup_jobs()
        assert len(self.project_a) == len(self.project_b)
        job_a0 = self.project_a.open_job({"a": 0})
        with open(job_a0.fn("test.txt"), "w") as file:
            file.write("newdata")
        with pytest.raises(FileSyncConflict):
            self.project_a.sync(self.project_b)
        with pytest.raises(FileSyncConflict):
            self.project_a.sync(self.project_b, selection=self.project_a)
        with pytest.raises(FileSyncConflict):
            self.project_a.sync(self.project_b, selection=self.project_b)
        assert len(self.project_a.find_jobs({"a": 0})) == 1
        assert len(self.project_b.find_jobs({"a": 0})) == 1
        with pytest.raises(FileSyncConflict):
            self.project_a.sync(
                self.project_b, selection=self.project_a.find_jobs({"a": 0})
            )
        with pytest.raises(FileSyncConflict):
            self.project_a.sync(
                self.project_b, selection=self.project_b.find_jobs({"a": 0})
            )
        with pytest.raises(FileSyncConflict):
            self.project_a.sync(
                self.project_b,
                selection=[job.id for job in self.project_a.find_jobs({"a": 0})],
            )
        with pytest.raises(FileSyncConflict):
            self.project_a.sync(
                self.project_b,
                selection=[job.id for job in self.project_b.find_jobs({"a": 0})],
            )
        f = {"a": {"$ne": 0}}
        self.project_a.sync(self.project_b, selection=self.project_a.find_jobs(f))
        self.project_a.sync(self.project_b, selection=self.project_b.find_jobs(f))
        self.project_a.sync(
            self.project_b, selection=[job.id for job in self.project_a.find_jobs(f)]
        )
        self.project_a.sync(
            self.project_b, selection=[job.id for job in self.project_b.find_jobs(f)]
        )



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Command line interface (CLI) for signac."""

import argparse
import atexit
import code
import difflib
import errno
import importlib
import json
import logging
import os
import platform
import re
import shutil
import sys
import warnings
from pprint import pformat, pprint
from rlcompleter import Completer

from tqdm import tqdm

try:
    import readline
except ImportError:
    READLINE = False
else:
    READLINE = True

from . import get_project, init_project
from ._config import (
    PROJECT_CONFIG_FN,
    USER_CONFIG_FN,
    _Config,
    _load_config,
    _locate_config_dir,
    _read_config_file,
)
from ._utility import _print_err, _query_yes_no, _safe_relpath
from ._vendor.configobj import Section, flatten_errors
from .diff import diff_jobs
from .errors import (
    DestinationExistsError,
    DocumentSyncConflict,
    FileSyncConflict,
    SchemaSyncConflict,
    SyncConflict,
)
from .filterparse import parse_filter_arg
from .import_export import _SchemaPathEvaluationError, export_jobs
from .sync import DocSync, FileSync
from .version import __version__

MSG_SYNC_SPECIFY_KEY = """
Synchronization conflict occurred, no strategy defined to synchronize keys:
{keys}

Use the `-k/--key` option to specify a regular expression pattern matching
all keys that should be overwritten, `--all-keys` to overwrite all conflicting
keys, or `--no-keys` to overwrite none of the conflicting keys."""


MSG_SYNC_FILE_CONFLICT = """
Synchronization conflict occurred, no strategy defined to synchronize files:
{files}

Use the `-s/--strategy` option to specify a file synchronization strategy,
or the `-u/--update` option to overwrite all files which have a more recent
modification time stamp."""


MSG_SYNC_STATS = """
Number of files transferred: {stats.num_files}
Total transfer volume:       {stats.volume}
"""


SHELL_BANNER = """Python {python_version}
signac {signac_version} 🎨

Project:\t{path}{job_banner}
Size:\t\t{size}

Interact with the project interface using the "project" or "pr" variable.
Type "help(project)" or "help(signac)" for more information."""


SHELL_BANNER_INTERACTIVE_IMPORT = (
    SHELL_BANNER
    + """

The data from origin '{origin}' has been imported into a temporary project.
Synchronize your project with the temporary project, for example with:

                    project.sync(tmp_project, recursive=True)
"""
)


SHELL_HISTORY_FN = os.sep.join((".signac", "shell_history"))

warnings.simplefilter("default")


def _fmt_bytes(nbytes, suffix="B"):
    """Format number of bytes.

    Adapted from: https://stackoverflow.com/a/1094933
    """
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(nbytes) < 1024.0:
            return f"{nbytes:3.1f} {unit}{suffix}"
        nbytes /= 1024.0
    return f"{nbytes:.1f} Yi{suffix}"


def _open_job_by_id(project, job_id):
    """Attempt to open a job by id and provide user feedback on error."""
    try:
        return project.open_job(id=job_id)
    except KeyError:
        close_matches = difflib.get_close_matches(
            job_id, [job.id[: len(job_id)] for job in project.find_jobs()]
        )
        msg = f"Did not find job corresponding to id '{job_id}'."
        if len(close_matches) == 1:
            msg += f" Did you mean '{close_matches[0]}'?"
        elif len(close_matches) > 1:
            msg += " Did you mean any of [{}]?".format("|".join(close_matches))
        raise KeyError(msg)
    except LookupError:
        n = project.min_len_unique_id()
        raise LookupError(
            "Multiple matches for abbreviated id '{}'. "
            "Use at least {} characters for guaranteed "
            "unique ids.".format(job_id, n)
        )


def _find_with_filter_or_none(args):
    """Return a filtered subset of jobs or None."""
    if args.job_id or args.filter:
        return _find_with_filter(args)
    else:
        return None


def _find_with_filter(args):
    """Return a filtered subset of jobs."""
    if getattr(args, "job_id", None):
        if args.filter:
            raise ValueError("Can't provide both 'job-id' and filter arguments!")
        else:
            return args.job_id

    project = get_project()
    filter_ = parse_filter_arg(args.filter) or {}
    if not filter_:
        filter_ = None
    return project._find_job_ids(filter=filter_)


def main_job(args):
    """Handle job subcommand."""
    project = get_project()
    if args.statepoint == "-":
        sp = input()
    else:
        sp = args.statepoint
    try:
        statepoint = json.loads(sp)
    except ValueError:
        _print_err(f"Error while reading state point: '{sp}'")
        raise
    job = project.open_job(statepoint)
    if args.create:
        job.init()
    if args.path:
        print(job.path)
    else:
        print(job)


def main_statepoint(args):
    """Handle statepoint subcommand."""
    project = get_project()
    if args.job_id:
        jobs = (_open_job_by_id(project, jid) for jid in args.job_id)
    else:
        jobs = project
    for job in jobs:
        if args.pretty:
            pprint(job.statepoint(), depth=args.pretty)
        else:
            print(json.dumps(job.statepoint(), indent=args.indent, sort_keys=args.sort))


def main_document(args):
    """Handle document subcommand."""
    project = get_project()
    for job_id in _find_with_filter(args):
        job = _open_job_by_id(project, job_id)
        if args.pretty:
            pprint(job.document(), depth=args.pretty)
        else:
            print(json.dumps(job.document(), indent=args.indent, sort_keys=args.sort))


def main_remove(args):
    """Handle remove subcommand."""
    project = get_project()
    for job_id in args.job_id:
        job = _open_job_by_id(project, job_id)
        if args.interactive and not _query_yes_no(
            "Are you sure you want to {action} job with id '{job.id}'?".format(
                action="clear" if args.clear else "remove", job=job
            ),
            default="no",
        ):
            continue
        if args.clear:
            job.clear()
            logging.info("Cleared f{job_id}")
        else:
            job.remove()
            logging.info("Removing f{job_id}")


def main_move(args):
    """Handle move subcommand."""
    project = get_project()
    dst_project = get_project(path=args.project)
    for job_id in args.job_id:
        try:
            job = _open_job_by_id(project, job_id)
            job.move(dst_project)
        except DestinationExistsError:
            _print_err(f"Destination already exists: '{job}' in '{dst_project}'.")
        else:
            _print_err(f"Moved '{job}' to '{dst_project}'.")


def main_clone(args):
    """Handle clone subcommand."""
    project = get_project()
    dst_project = get_project(path=args.project)
    for job_id in args.job_id:
        try:
            job = _open_job_by_id(project, job_id)
            dst_project.clone(job)
        except DestinationExistsError:
            _print_err(f"Destination already exists: '{job}' in '{dst_project}'.")
        else:
            _print_err(f"Cloned '{job}' to '{dst_project}'.")


def main_find(args):
    """Handle find subcommand."""
    project = get_project()

    len_id = 6
    if args.one_line:
        # Only get the project's minimum length of unique id if it is needed.
        len_id = max(len_id, project.min_len_unique_id())

    # --show = --sp --doc --pretty 3
    # if --sp or --doc are also specified, those subsets of keys will be used
    if args.show:
        if args.sp is None:
            args.sp = []
        if args.doc is None:
            args.doc = []

    def format_lines(cat, id_, s):
        if args.one_line:
            if isinstance(s, dict):
                s = json.dumps(s, sort_keys=True)
            return f"{id_[:len_id]} {cat}\t{s}"
        else:
            return pformat(s, depth=args.pretty)

    try:
        for job_id in _find_with_filter(args):
            print(job_id)
            job = project.open_job(id=job_id)

            if args.sp is not None:
                statepoint = job.statepoint()
                if len(args.sp) != 0:
                    statepoint = {
                        key: statepoint[key] for key in args.sp if key in statepoint
                    }
                print(format_lines("sp ", job_id, statepoint))

            if args.doc is not None:
                doc = job.document()
                if len(args.doc) != 0:
                    doc = {key: doc[key] for key in args.doc if key in doc}
                print(format_lines("sp ", job_id, doc))
    except OSError as error:
        if error.errno == errno.EPIPE:
            sys.stderr.close()
        else:
            raise


def main_diff(args):
    """Handle diff subcommand."""
    project = get_project()

    jobs = _find_with_filter_or_none(args)
    jobs = (
        (_open_job_by_id(project, job) for job in jobs) if jobs is not None else project
    )

    diff = diff_jobs(*jobs)

    for job_id, statepoint in diff.items():
        print(job_id)
        pprint(statepoint)


def main_view(args):
    """Handle view subcommand."""
    project = get_project()
    project.create_linked_view(
        prefix=args.prefix,
        path=args.path,
        job_ids=_find_with_filter(args),
    )


def main_init(args):
    """Handle init subcommand."""
    init_project(path=os.getcwd())
    _print_err("Initialized project.")


def main_schema(args):
    """Handle schema subcommand."""
    project = get_project()
    print(
        project.detect_schema(
            exclude_const=args.exclude_const, subset=_find_with_filter_or_none(args)
        ).format(
            depth=args.depth, precision=args.precision, max_num_range=args.max_num_range
        )
    )


def main_sync(args):
    """Handle sync subcommand."""
    # Validate provided argument combinations
    if args.archive:
        args.recursive = True
        args.links = True
        args.times = True
        args.perms = True
        args.owner = True
        args.group = True

    if args.update:
        if args.strategy is not None:
            raise ValueError(
                "Can't provide both the '-u/--update' and a '-s/--strategy argument!"
            )
        args.strategy = "update"

    if args.times and not args.perms:
        raise NotImplementedError(
            "The '-t/--times' option can only be used in combination with the "
            "'-p/--perms argument."
        )

    if args.size_only or args.round_times:
        # Apply monkey patch
        import filecmp
        import stat

        if args.size_only:

            def _sig(st):
                return (stat.S_IFMT(st.st_mode), st.st_size)

        else:

            def _sig(st):
                return (stat.S_IFMT(st.st_mode), st.st_size, int(st.st_mtime))

        filecmp._sig = _sig

    # Setup synchronization process

    source = get_project(path=args.source)
    try:
        destination = get_project(path=args.destination)
    except LookupError:
        _print_err("WARNING: The destination does not appear to be a project path.")
        raise
    selection = _find_with_filter_or_none(args)

    if args.strategy:
        if args.strategy[0].isupper():
            strategy = getattr(FileSync, args.strategy)()
        else:
            strategy = getattr(FileSync, args.strategy)
    else:
        strategy = None

    if sum((args.all_keys, args.no_keys, args.key is not None)) > 1:
        raise ValueError("You can only provide one key argument!")
    elif args.all_keys:
        doc_sync = DocSync.ByKey(lambda key: True)
    elif args.no_keys:
        doc_sync = DocSync.ByKey(lambda key: False)
    elif args.key:
        try:
            re.compile(args.key)
        except re.error as e:
            raise RuntimeError(f"Illegal regular expression '{args.key}': '{e}'.")
        doc_sync = DocSync.ByKey(lambda key: re.match(args.key, key))
    else:
        doc_sync = DocSync.ByKey()

    try:
        _print_err(f"Synchronizing '{source}' -> '{destination}'...")
        stats = destination.sync(
            other=source,
            strategy=strategy,
            recursive=args.recursive,
            follow_symlinks=not args.links,
            preserve_permissions=args.perms,
            preserve_times=args.times,
            preserve_owner=args.owner,
            preserve_group=args.group,
            exclude=args.exclude,
            doc_sync=doc_sync,
            selection=selection,
            check_schema=not (args.merge or args.force),
            dry_run=args.dry_run,
            parallel=args.parallel,
            deep=args.deep,
            collect_stats=args.stats,
        )
        if stats is not None:
            if args.human_readable:
                stats = stats._replace(volume=_fmt_bytes(stats.volume))
            print("\n# Transfer statistics", "(dry run)" if args.dry_run else "")
            if args.json:
                print(json.dumps(stats._asdict()))
            else:
                print(MSG_SYNC_STATS.format(stats=stats))
    except SchemaSyncConflict as error:
        _print_err(
            "Synchronizing two projects with different schema requires the -m/--merge option."
        )
        diff_src = error.schema_src.difference(error.schema_dst)
        diff_dst = error.schema_dst.difference(error.schema_src)
        only_in_dst = diff_dst.difference(diff_src)
        only_in_src = diff_src.difference(diff_dst)
        diff_value = diff_src.intersection(diff_dst)
        if only_in_src:
            _print_err(
                "Keys found only in the source schema: {}".format(
                    ", ".join(only_in_src)
                )
            )
        if only_in_dst:
            _print_err(
                "Keys found only in the destination schema: {}".format(
                    ", ".join(only_in_dst)
                )
            )
        if diff_value:
            _print_err(
                "Keys having different values in source and destination: {}".format(
                    ", ".join(diff_value)
                )
            )
    except DocumentSyncConflict as error:
        _print_err(MSG_SYNC_SPECIFY_KEY.format(keys=", ".join(error.keys)))
    except FileSyncConflict as error:
        _print_err(MSG_SYNC_FILE_CONFLICT.format(files=error))
    else:
        if doc_sync.skipped_keys:
            _print_err("Skipped key(s):", ", ".join(sorted(doc_sync.skipped_keys)))
        _print_err("Done.")
        return
    raise RuntimeWarning("Synchronization aborted.")


def _main_import_interactive(project, origin, args):
    from .import_export import _prepare_import_into_project

    if args.move:
        raise ValueError(
            "Cannot use '--move' in combination with '--sync-interactive'."
        )

    with project.temporary_project() as tmp_project:
        _print_err("Prepare data space for import...")
        with _prepare_import_into_project(
            origin, tmp_project, args.schema_path
        ) as data_mapping:
            paths = {}
            for src, copy_executor in tqdm(
                dict(data_mapping).items(), desc="Import to temporary project"
            ):
                paths[src] = copy_executor()

            local_ns = dict(
                signac=importlib.import_module(__package__),
                project=project,
                pr=project,
                tmp_project=tmp_project,
            )
            if READLINE:
                readline.set_completer(Completer(local_ns).complete)
                readline.parse_and_bind("tab: complete")
            code.interact(
                local=local_ns,
                banner=SHELL_BANNER_INTERACTIVE_IMPORT.format(
                    python_version=sys.version,
                    signac_version=__version__,
                    job_banner="",
                    path=project.path,
                    size=len(project),
                    origin=args.origin,
                ),
            )

            return paths


def _main_import_non_interactive(project, origin, args):
    from .import_export import _prepare_import_into_project

    try:
        paths = {}
        if args.sync:
            with project.temporary_project() as tmp_project:
                _print_err("Prepare data space for import...")
                with _prepare_import_into_project(
                    origin, tmp_project, args.schema_path
                ) as mapping:
                    for src, copy_executor in tqdm(
                        dict(mapping).items(), desc="Import to temporary project"
                    ):
                        paths[src] = copy_executor()
                    _print_err("Synchronizing project with temporary project...")
                    project.sync(tmp_project, recursive=True)
        else:
            _print_err("Prepare data space for import...")
            with _prepare_import_into_project(
                origin, project, args.schema_path
            ) as data_mapping:
                for src, copy_executor in tqdm(dict(data_mapping).items(), "Importing"):
                    paths[src] = copy_executor(
                        copytree=shutil.move if args.move else None
                    )
    except DestinationExistsError as error:
        _print_err(f"Destination '{error.destination}' already exists.")
        if not args.sync:
            _print_err("Consider using '--sync' or '--sync-interactive'!")
    except SyncConflict as error:
        _print_err(f"Synchronization failed with error: {error}")
        _print_err("Consider using '--sync-interactive'!")
    else:
        return paths


def main_import(args):
    """Handle import subcommand."""
    if args.move and os.path.isfile(args.origin):
        raise ValueError("Cannot use '--move' when importing from a file.")
    if args.move and (args.sync or args.sync_interactive):
        raise ValueError(
            "Cannot use '--move' in combination with '--sync' or '--sync-interactive'."
        )

    project = get_project()
    if args.sync_interactive:
        paths = _main_import_interactive(project, args.origin, args)
    else:
        paths = _main_import_non_interactive(project, args.origin, args)

    if paths is None:
        _print_err("Import failed.")
    elif len(paths):
        _print_err(f"Imported {len(paths)} job(s).")
    elif paths is not None:
        _print_err("Nothing to import.")


def main_export(args):
    """Handle export subcommand."""
    if args.move and os.path.splitext(args.target)[1] != "":
        raise RuntimeError(
            "The '--move' argument can only be used when exporting to directories."
        )
    copytree = shutil.move if args.move else None

    project = get_project()
    jobs = [project.open_job(id=job_id) for job_id in _find_with_filter(args)]

    paths = {}
    with tqdm(total=len(jobs), desc="Export") as pbar:
        try:
            for src, dst in export_jobs(
                jobs=jobs, target=args.target, path=args.schema_path, copytree=copytree
            ):
                paths[src] = dst
                pbar.update(1)
        except _SchemaPathEvaluationError as error:
            raise RuntimeWarning(
                f"An error occurred while evaluating the schema path: {error}"
            )

    if paths:
        _print_err(f"Exported {len(paths)} job(s).")
    else:
        _print_err("No jobs to export.")


def main_update_cache(args):
    """Handle update-cache subcommand."""
    project = get_project()
    _print_err("Updating cache...")
    n = project.update_cache()
    if n is None:
        _print_err("Cache is up to date.")
    else:
        _print_err(f"Updated cache (size={n}).")


def main_migrate(args):
    """Migrate the project's schema to the current schema version."""
    from .migration import _get_config_schema_version, apply_migrations
    from .version import SCHEMA_VERSION

    root = args.root_directory if args.root_directory else os.getcwd()

    schema_version = int(SCHEMA_VERSION)
    config_schema_version = _get_config_schema_version(root, schema_version)

    if config_schema_version > schema_version:
        _print_err(
            f"The schema version of the project ({config_schema_version}) is "
            "newer than the schema version {schema_version} supported by "
            f"signac version {__version__}. Try updating signac."
        )
    elif config_schema_version == schema_version:
        _print_err(
            f"The schema version of the project ({config_schema_version}) is "
            "up to date. Nothing to do."
        )
    elif args.yes or _query_yes_no(
        "Do you want to migrate this project's schema version from "
        f"'{config_schema_version}' to '{schema_version}'? WARNING: THIS "
        "PROCESS IS IRREVERSIBLE!",
        "no",
    ):
        apply_migrations(root)


def main_config_show(args):
    """Handle config show subcommand."""
    cfg = None
    if args.local and args.globalcfg:
        raise ValueError("You can specify either -l/--local or -g/--global, not both.")
    elif args.local:
        if os.path.isfile(PROJECT_CONFIG_FN):
            cfg = _read_config_file(PROJECT_CONFIG_FN)
    elif args.globalcfg:
        cfg = _read_config_file(USER_CONFIG_FN)
    else:
        cfg = _load_config(_locate_config_dir(os.getcwd()))
    if not cfg:
        if args.local:
            mode = "local"
        elif args.globalcfg:
            mode = "global"
        else:
            mode = "local or global"
        _print_err(f"Did not find a {mode} configuration file.")
        return
    for key in args.key:
        for kt in key.split("."):
            cfg = cfg.get(kt)
            if cfg is None:
                break
    if not isinstance(cfg, Section):
        print(cfg)
    else:
        for line in _Config(cfg).write():
            print(line)


def _verify_config(cfg, preserve_errors=True):
    """Verify provided configuration."""
    verification = cfg.verify(preserve_errors=preserve_errors)
    if verification is True:
        _print_err("Passed.")
    else:
        for entry in flatten_errors(cfg, verification):
            # each entry is a tuple
            section_list, key, error = entry
            if key is not None:
                section_list.append(key)
            else:
                section_list.append("[missing section]")
            section_string = ".".join(section_list)
            if error is False:
                error = "Possibly invalid or missing."
            else:
                error = type(error).__name__
            _print_err(" ".join((section_string, ":", error)))


def main_config_verify(args):
    """Handle config verify subcommand."""
    cfg = None
    if args.local and args.globalcfg:
        raise ValueError("You can specify either -l/--local or -g/--global, not both.")
    elif args.local:
        if os.path.isfile(PROJECT_CONFIG_FN):
            cfg = _read_config_file(PROJECT_CONFIG_FN)
    elif args.globalcfg:
        cfg = _read_config_file(USER_CONFIG_FN)
    else:
        cfg = _load_config(_locate_config_dir(os.getcwd()))
    if not cfg:
        if args.local:
            mode = "local"
        elif args.globalcfg:
            mode = "global"
        else:
            mode = "local or global"
        _print_err(f"Did not find a {mode} configuration file.")
        return
    else:
        if cfg.filename is not None:
            _print_err(f"Verification of config file '{cfg.filename}'.")
        _verify_config(cfg)


def main_config_set(args):
    """Handle config set subcommand."""
    if not (args.local or args.globalcfg):
        args.local = True
    fn_config = None
    if args.local and args.globalcfg:
        raise ValueError("You can specify either -l/--local or -g/--global, not both.")
    elif args.local:
        if os.path.isfile(PROJECT_CONFIG_FN):
            fn_config = PROJECT_CONFIG_FN
    elif args.globalcfg:
        fn_config = USER_CONFIG_FN
    else:
        raise ValueError(
            "You need to specify either -l/--local or -g/--global "
            "to specify which configuration to modify."
        )
    cfg = _read_config_file(fn_config)
    keys = args.key.split(".")
    if len(args.value) == 0:
        raise ValueError("No value argument provided!")
    elif len(args.value) == 1:
        args.value = args.value[0]
    sec = cfg
    for key in keys[:-1]:
        sec = sec.setdefault(key, {})
    try:
        sec[keys[-1]] = args.value
        _print_err(f"Updated value '{args.key}'='{args.value}'.")
    except TypeError:
        raise KeyError(args.key)
    _print_err(f"Writing configuration to '{os.path.abspath(fn_config)}'.")
    cfg.write()


def main_shell(args):
    """Handle shell subcommand."""
    if args.file and args.command:
        raise ValueError(
            "Cannot provide file and -c/--command argument at the same time!"
        )

    try:
        project = get_project()
    except LookupError:
        print("signac", __version__)
        print("No project within this directory.")
        print("If you want to initialize a project, execute `$ signac init`.")
    else:
        _jobs = _find_with_filter(args)

        def jobs():
            for id_ in _jobs:
                yield project.open_job(id=id_)

        if len(_jobs) == 1:
            job = _open_job_by_id(project, list(_jobs)[0])
        else:
            try:
                job = project.get_job()
            except LookupError:
                job = None

        local_ns = dict(
            project=project,
            pr=project,
            jobs=iter(jobs()),
            job=job,
            signac=sys.modules["signac"],
        )

        if args.file or args.command:
            interpreter = code.InteractiveInterpreter(locals=local_ns)
            if args.file and args.file == "-":
                try:
                    while True:
                        interpreter.runsource(
                            input(), filename="<input>", symbol="exec"
                        )
                except EOFError:
                    pass
            elif args.file:
                with open(args.file) as file:
                    interpreter.runsource(
                        file.read(), filename=args.file, symbol="exec"
                    )
            else:
                interpreter.runsource(args.command, filename="<input>", symbol="exec")
        else:  # interactive
            if READLINE:
                if "PyPy" not in platform.python_implementation():
                    fn_hist = project.fn(SHELL_HISTORY_FN)
                    try:
                        readline.read_history_file(fn_hist)
                        readline.set_history_length(1000)
                    except FileNotFoundError:
                        pass
                    except PermissionError:
                        print(
                            "Warning: Shell history could not be read from "
                            "{}.".format(_safe_relpath(fn_hist))
                        )

                    def write_history_file():
                        try:
                            readline.write_history_file(fn_hist)
                        except PermissionError:
                            print(
                                "Warning: Shell history could not be written to "
                                "{}.".format(_safe_relpath(fn_hist))
                            )

                    atexit.register(write_history_file)
                readline.set_completer(Completer(local_ns).complete)
                readline.parse_and_bind("tab: complete")
            code.interact(
                local=local_ns,
                banner=SHELL_BANNER.format(
                    python_version=sys.version,
                    signac_version=__version__,
                    job_banner=f"\nJob:\t\t{job.id}" if job is not None else "",
                    path=project.path,
                    size=len(project),
                ),
            )


def main():
    """Provide command line interface."""
    parser = argparse.ArgumentParser(
        description="signac aids in the management, access and analysis of "
        "large-scale computational investigations."
    )
    parser.add_argument(
        "--debug", action="store_true", help="Show traceback on error for debugging."
    )
    parser.add_argument(
        "--version", action="store_true", help="Display the version number and exit."
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        help="Set level of verbosity.",
        action="count",
        default=2,
    )
    parser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        help="Answer all questions with yes. Useful for scripted interaction.",
    )
    subparsers = parser.add_subparsers()

    parser_init = subparsers.add_parser("init")
    parser_init.set_defaults(func=main_init)

    parser_job = subparsers.add_parser("job")
    parser_job.add_argument(
        "statepoint",
        nargs="?",
        default="-",
        type=str,
        help="The job's state point in JSON format. Omit this argument to read from STDIN.",
    )
    parser_job.add_argument(
        "-w",
        "--workspace",
        action="store_true",
        help="Print the job's workspace path instead of the job id.",
    )
    parser_job.add_argument(
        "-p",
        "--path",
        action="store_true",
        help="Print the job's path instead of the job id.",
    )
    parser_job.add_argument(
        "-c",
        "--create",
        action="store_true",
        help="Create the job directory if necessary.",
    )
    parser_job.set_defaults(func=main_job)

    parser_statepoint = subparsers.add_parser(
        "statepoint",
        description="Print the state point(s) corresponding to one or more job ids.",
    )
    parser_statepoint.add_argument(
        "job_id",
        nargs="*",
        type=str,
        help="One or more job ids. The corresponding jobs must be initialized.",
    )
    parser_statepoint.add_argument(
        "-p",
        "--pretty",
        type=int,
        nargs="?",
        const=3,
        help="Print state point in pretty format. "
        "An optional argument to this flag specifies the maximal "
        "depth a state point is printed.",
    )
    parser_statepoint.add_argument(
        "-i",
        "--indent",
        type=int,
        nargs="?",
        const="2",
        help="Specify the indentation of the JSON formatted state point.",
    )
    parser_statepoint.add_argument(
        "-s",
        "--sort",
        action="store_true",
        help="Sort the state point keys for output.",
    )
    parser_statepoint.set_defaults(func=main_statepoint)

    parser_diff = subparsers.add_parser(
        "diff", description="Find the difference among job state points."
    )
    parser_diff.add_argument(
        "job_id",
        nargs="*",
        type=str,
        help="One or more job ids. The corresponding jobs must be initialized.",
    )
    parser_diff.add_argument(
        "-p",
        "--pretty",
        type=int,
        nargs="?",
        const=3,
        help="Print state point in pretty format. "
        "An optional argument to this flag specifies the maximal "
        "depth a state point is printed.",
    )
    parser_diff.add_argument(
        "-i",
        "--indent",
        type=int,
        nargs="?",
        const="2",
        help="Specify the indentation of the JSON formatted state point.",
    )
    parser_diff.add_argument(
        "-f",
        "--filter",
        type=str,
        nargs="+",
        help="Limit the diff to jobs matching the filter.",
    )
    parser_diff.set_defaults(func=main_diff)

    parser_document = subparsers.add_parser(
        "document",
        description="Print the document(s) corresponding to one or more job ids.",
    )
    parser_document.add_argument(
        "job_id",
        nargs="*",
        type=str,
        help="One or more job ids. The job corresponding to a job id must be initialized.",
    )
    parser_document.add_argument(
        "-p",
        "--pretty",
        type=int,
        nargs="?",
        const=3,
        help="Print document in pretty format. "
        "An optional argument to this flag specifies the maximal "
        "depth a document is printed.",
    )
    parser_document.add_argument(
        "-i",
        "--indent",
        type=int,
        nargs="?",
        const="2",
        help="Specify the indentation of the JSON formatted state point.",
    )
    parser_document.add_argument(
        "-s",
        "--sort",
        action="store_true",
        help="Sort the document keys for output in JSON format.",
    )
    parser_document.add_argument(
        "-f",
        "--filter",
        type=str,
        nargs="+",
        help="Show documents of jobs matching the filter.",
    )
    parser_document.set_defaults(func=main_document)

    parser_remove = subparsers.add_parser("rm")
    parser_remove.add_argument(
        "job_id", type=str, nargs="+", help="One or more job ids of jobs to remove."
    )
    parser_remove.add_argument(
        "-c",
        "--clear",
        action="store_true",
        help="Do not completely remove, but only clear the job.",
    )
    parser_remove.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        help="Request confirmation before attempting to remove/clear each job.",
    )
    parser_remove.set_defaults(func=main_remove)

    parser_move = subparsers.add_parser("move")
    parser_move.add_argument(
        "project",
        type=str,
        help="The destination project directory.",
    )
    parser_move.add_argument(
        "job_id",
        nargs="+",
        type=str,
        help="One or more job ids. The corresponding jobs must be initialized.",
    )
    parser_move.set_defaults(func=main_move)

    parser_clone = subparsers.add_parser("clone")
    parser_clone.add_argument(
        "project",
        type=str,
        help="The directory of the project to clone one or more jobs in.",
    )
    parser_clone.add_argument(
        "job_id",
        nargs="+",
        type=str,
        help="One or more job ids. The corresponding jobs must be initialized.",
    )
    parser_clone.set_defaults(func=main_clone)

    parser_find = subparsers.add_parser(
        "find",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="All filter arguments may be provided either directly in "
        "JSON encoding or in a simplified form using the query API. For "
        "example, the following two bash commands are equivalent:\n\n"
        "$ signac find key 42\n"
        "$ signac find '{\"a\": 42}'\n\n"
        "For the full syntax, see:\n"
        "https://docs.signac.io/en/latest/query.html#simplified-filter",
    )
    parser_find.add_argument(
        "filter",
        type=str,
        nargs="*",
        help="Find jobs matching the filter.",
    )
    parser_find.add_argument(
        "-s",
        "--show",
        type=int,
        nargs="?",
        const=3,
        help="Show the state point and document of each job. Equivalent to "
        "--sp --doc --pretty 3.",
    )
    parser_find.add_argument(
        "--sp",
        type=str,
        nargs="*",
        help="Show the state point of each job. Can be passed the list of "
        "state point keys to print (if they exist for a given job).",
    )
    parser_find.add_argument(
        "--doc",
        type=str,
        nargs="*",
        help="Show the document of each job. Can be passed the list of "
        "document keys to print (if they exist for a given job).",
    )
    parser_find.add_argument(
        "-p",
        "--pretty",
        type=int,
        nargs="?",
        const=3,
        default=3,
        help="Pretty print output when using --sp, --doc, or ---show. "
        "Argument is the depth to which keys are printed.",
    )
    parser_find.add_argument(
        "-1",
        "--one-line",
        action="store_true",
        help="Print output in JSON and on one line.",
    )
    parser_find.set_defaults(func=main_find)

    parser_view = subparsers.add_parser(
        "view",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Generate a human readable set of paths representing "
        "state points in the workspace, e.g. "
        "view/param_name_1/param_value_1/param_name_2/param_value_2/job.\n\n"
        "The leaf nodes of this directory structure are symlinks (named "
        '"job") into the corresponding job directory.\n\n'
        "The jobs can be filtered using the query API used by the find "
        "command. For the full syntax, see:\n"
        "https://docs.signac.io/en/latest/query.html#simplified-filter\n\n"
        "All positional arguments must be provided before any keyword "
        "arguments. In particular, the prefix and path must be "
        "specified before arguments such as the filters, like in the "
        "following bash command:\n\n"
        "$ signac view $PREFIX $VIEW_PATH -f $FILTERS -d $DOC_FILTERS.",
    )
    parser_view.add_argument(
        "-p",
        "--prefix",
        type=str,
        nargs="?",
        default="view",
        help="The path where the view is to be created. Defaults to view.",
    )
    parser_view.add_argument(
        "path",
        type=str,
        nargs="?",
        default="{{auto}}",
        help="The path used for the generation of the linked view hierarchy, "
        "defaults to '{{auto}}' (see Project.export_to for information "
        "on how this is expanded).",
    )
    selection_group = parser_view.add_argument_group("select")
    selection_group.add_argument(
        "-f",
        "--filter",
        type=str,
        nargs="+",
        help="Limit the view to jobs matching the filter.",
    )
    selection_group.add_argument(
        "-j",
        "--job-id",
        type=str,
        nargs="+",
        help="Limit the view to jobs with these job ids.",
    )
    parser_view.set_defaults(func=main_view)

    parser_schema = subparsers.add_parser("schema")
    parser_schema.add_argument(
        "-x",
        "--exclude-const",
        action="store_true",
        help="Exclude state point parameters, which are constant over the "
        "complete project data space.",
    )
    parser_schema.add_argument(
        "-t",
        "--depth",
        type=int,
        default=0,
        help="A non-zero value will format the schema in a nested representation "
        "up to the specified depth. The default is a flat view (depth=0).",
    )
    parser_schema.add_argument(
        "-p",
        "--precision",
        type=int,
        help="Round all numerical values up to the given precision.",
    )
    parser_schema.add_argument(
        "-r",
        "--max-num-range",
        type=int,
        default=5,
        help="The maximum number of entries shown for a value range, defaults to 5.",
    )
    selection_group = parser_schema.add_argument_group("select")
    selection_group.add_argument(
        "-f",
        "--filter",
        type=str,
        nargs="+",
        help="Detect schema only for jobs matching the filter.",
    )
    selection_group.add_argument(
        "-j",
        "--job-id",
        type=str,
        nargs="+",
        help="Detect schema only for jobs with the given job ids.",
    )
    parser_schema.set_defaults(func=main_schema)

    parser_shell = subparsers.add_parser("shell")
    parser_shell.add_argument(
        "file", type=str, nargs="?", help="Execute Python script in file."
    )
    parser_shell.add_argument(
        "-c", "--command", type=str, help="Execute Python program passed as string."
    )
    selection_group = parser_shell.add_argument_group(
        "select",
        description="Specify one or more jobs to preset the `jobs` variable as a generator "
        "over all job handles associated with the given selection. If the selection "
        "contains only one job, an additional `job` variable is referencing that "
        "single job, otherwise it is `None`.",
    )
    selection_group.add_argument(
        "-f",
        "--filter",
        type=str,
        nargs="+",
        help="Reduce selection to jobs matching the filter.",
    )
    selection_group.add_argument(
        "-j",
        "--job-id",
        type=str,
        nargs="+",
        help="Reduce selection to jobs matching the given job ids.",
    )
    parser_shell.set_defaults(func=main_shell)

    parser_sync = subparsers.add_parser(
        "sync",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Use this command to synchronize this project with "
        "another project, similar to the synchronization of two directories "
        "with `rsync`. Data is always copied from the source to the "
        "destination.\n\n"
        "For example, the following bash command will synchronize all jobs "
        "within this project with jobs in the other project, overwrite files "
        "if the source files are newer, and overwrite all conflicting keys in "
        "the project and job documents:\n\n"
        "$ signac sync /path/to/other/project --update --all-keys",
    )
    parser_sync.add_argument(
        "source",
        help="The directory of the project that this project should be synchronized with.",
    )
    parser_sync.add_argument(
        "destination",
        nargs="?",
        help="Optional: The directory of the project that should be modified for "
        "synchronization, defaults to the local project.",
    )

    sync_group = parser_sync.add_argument_group("copy options")
    sync_group.add_argument(
        "-a",
        "--archive",
        action="store_true",
        help="archive mode; equivalent to: '-rltpog'",
    )
    sync_group.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="Do not skip sub-directories, but synchronize recursively.",
    )
    sync_group.add_argument(
        "-l",
        "--links",
        action="store_true",
        help="Copy symbolic links as symbolic links pointing to the original source.",
    )
    sync_group.add_argument(
        "-p", "--perms", action="store_true", help="Preserve permissions."
    )
    sync_group.add_argument(
        "-o", "--owner", action="store_true", help="Preserve owner."
    )
    sync_group.add_argument(
        "-g", "--group", action="store_true", help="Preserve group."
    )
    sync_group.add_argument(
        "-t",
        "--times",
        action="store_true",
        help="Preserve file modification times (requires -p).",
    )
    sync_group.add_argument(
        "-x",
        "--exclude",
        type=str,
        nargs="?",
        const=".*",
        help="Exclude all files matching the given pattern. Exclude all files "
        "if this option is provided without any argument.",
    )
    sync_group.add_argument(
        "-I",
        "--ignore-times",
        action="store_true",
        dest="deep",
        help="Never rely on file metadata such as the size or the modification time "
        "when determining file differences.",
    )
    sync_group.add_argument(
        "--size-only",
        action="store_true",
        help="Ignore modification times during file comparison. Useful when synchronizing "
        "between file systems with different timestamp resolution.",
    )
    sync_group.add_argument(
        "--round-times",
        action="store_true",
        help="Round modification times during file comparison. Useful when synchronizing "
        "between file systems with different timestamp resolution.",
    )
    sync_group.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        help="Do not actually execute the synchronization. Increase the output verbosity "
        "to see messages about what would potentially happen.",
    )
    sync_group.add_argument(
        "-u",
        "--update",
        action="store_true",
        help="Skip files with newer modification time stamp."
        "This is a short-cut for: --strategy=update.",
    )

    strategy_group = parser_sync.add_argument_group("sync strategy")
    strategy_group.add_argument(
        "-s",
        "--strategy",
        type=str,
        choices=FileSync.keys(),
        help="Specify a synchronization strategy, for differing files.",
    )
    strategy_group.add_argument(
        "-k",
        "--key",
        type=str,
        help="Specify a regular expression for keys that should be overwritten "
        "as part of the project and job document synchronization.",
    )
    strategy_group.add_argument(
        "--all-keys",
        action="store_true",
        help="Overwrite all conflicting keys. Equivalent to `--key='.*'`.",
    )
    strategy_group.add_argument(
        "--no-keys", action="store_true", help="Never overwrite any conflicting keys."
    )

    parser_sync.add_argument(
        "--force", action="store_true", help="Ignore all warnings, just synchronize."
    )
    parser_sync.add_argument(
        "-m",
        "--merge",
        action="store_true",
        help="Clone all the jobs that are not present in destination from source.",
    )
    parser_sync.add_argument(
        "--parallel",
        type=int,
        nargs="?",
        const=True,
        help="Use multiple threads for synchronization."
        "You may optionally specify how many threads to "
        "use, otherwise all available processing units will be utilized.",
    )
    parser_sync.add_argument(
        "--stats", action="store_true", help="Provide file transfer statistics."
    )
    parser_sync.add_argument(
        "-H",
        "--human-readable",
        action="store_true",
        help="Provide statistics with human readable formatting.",
    )
    parser_sync.add_argument(
        "--json", action="store_true", help="Print statistics in JSON formatting."
    )

    selection_group = parser_sync.add_argument_group("select")
    selection_group.add_argument(
        "-f",
        "--filter",
        type=str,
        nargs="+",
        help="Only synchronize jobs matching the filter.",
    )
    selection_group.add_argument(
        "-j",
        "--job-id",
        type=str,
        nargs="+",
        help="Only synchronize jobs with the given job ids.",
    )
    parser_sync.set_defaults(func=main_sync)

    parser_import = subparsers.add_parser(
        "import",
        description="""Import an existing dataset into this project. Optionally provide a file path
 based schema to specify the state point metadata. Providing a path based schema is only necessary
 if the data set was not previously exported from a signac project.""",
    )
    parser_import.add_argument(
        "origin",
        default=".",
        nargs="?",
        help="The origin to import from. May be a path to a directory, a zipfile, or a tarball. "
        "Defaults to the current working directory.",
    )
    parser_import.add_argument(
        "schema_path",
        nargs="?",
        help="Specify an optional import path, such as 'foo/{foo:int}'. Possible type definitions "
        "include bool, int, float, and str. The type is assumed to be 'str' if no type is "
        "specified.",
    )
    parser_import.add_argument(
        "--move",
        action="store_true",
        help="Move the data upon import instead of copying. Can only be used when importing from "
        "a directory.",
    )
    parser_import.add_argument(
        "--sync",
        action="store_true",
        help="Attempt recursive synchronization with default arguments.",
    )
    parser_import.add_argument(
        "--sync-interactive",
        action="store_true",
        help="Synchronize the project with the origin data space interactively.",
    )
    parser_import.set_defaults(func=main_import)

    parser_export = subparsers.add_parser(
        "export",
        description="""Export the project data space (or a subset) to a directory, a zipfile,
 or a tarball.""",
    )
    parser_export.add_argument(
        "target",
        help="The target to export to. May be a path to a directory, a zipfile, or a tarball.",
    )
    parser_export.add_argument(
        "schema_path",
        nargs="?",
        help="Specify an optional export path, based on the job state point, e.g., "
        "'foo/{job.sp.foo}'.",
    )
    parser_export.add_argument(
        "--move",
        action="store_true",
        help="Move data to export target instead of copying. Can only be used when exporting "
        "to a directory target.",
    )
    selection_group = parser_export.add_argument_group("select")
    selection_group.add_argument(
        "-f",
        "--filter",
        type=str,
        nargs="+",
        help="Limit the jobs to export to those matching the filter.",
    )
    selection_group.add_argument(
        "-j",
        "--job-id",
        type=str,
        nargs="+",
        help="Limit the jobs to export to those matching the provided job ids.",
    )
    parser_export.set_defaults(func=main_export)

    parser_update_cache = subparsers.add_parser(
        "update-cache",
        description="Use this command to update the project's persistent state point cache.",
    )
    parser_update_cache.set_defaults(func=main_update_cache)

    parser_config = subparsers.add_parser("config")
    parser_config.add_argument(
        "-g",
        "--global",
        dest="globalcfg",
        action="store_true",
        help="Modify the global configuration.",
    )
    parser_config.add_argument(
        "-l", "--local", action="store_true", help="Modify the local configuration."
    )
    parser_config.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Skip sanity checks when modifying the configuration.",
    )
    config_subparsers = parser_config.add_subparsers()

    parser_show = config_subparsers.add_parser("show")
    parser_show.add_argument(
        "key",
        type=str,
        nargs="*",
        help="The key(s) to show, omit to show the full configuration.",
    )
    parser_show.set_defaults(func=main_config_show)

    parser_set = config_subparsers.add_parser("set")
    parser_set.add_argument("key", type=str, help="The key to modify.")
    parser_set.add_argument(
        "value", type=str, nargs="*", help="The value to set key to."
    )
    parser_set.add_argument(
        "-f", "--force", action="store_true", help="Override any validation warnings."
    )
    parser_set.set_defaults(func=main_config_set)

    parser_verify = config_subparsers.add_parser("verify")
    parser_verify.set_defaults(func=main_config_verify)

    parser_migrate = subparsers.add_parser(
        "migrate",
        description="Irreversibly migrate this project's schema version to the "
        "supported version.",
    )
    parser_migrate.add_argument(
        "-r",
        "--root-directory",
        type=str,
        default="",
        help="The path to the project.",
    )
    parser_migrate.add_argument(
        "-y",
        "--yes",
        action="store_true",
        help="Do not ask for confirmation.",
    )
    parser_migrate.set_defaults(func=main_migrate)

    # This is a hack, as argparse itself does not
    # allow to parse only --version without any
    # of the other required arguments.
    if "--version" in sys.argv:
        print("signac", __version__)
        sys.exit(0)

    args = parser.parse_args()
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        log_level = (
            logging.DEBUG
            if args.debug
            else [
                logging.CRITICAL,
                logging.ERROR,
                logging.WARNING,
                logging.INFO,
                logging.MORE,
                logging.DEBUG,
            ][min(args.verbosity, 5)]
        )
        logging.basicConfig(level=log_level)

    if not hasattr(args, "func"):
        parser.print_usage()
        sys.exit(2)
    try:
        args.func(args)
    except KeyboardInterrupt:
        _print_err("Interrupted.")
        if args.debug:
            raise
        sys.exit(1)
    except RuntimeWarning as warning:
        _print_err(f"Warning: {warning}")
        if args.debug:
            raise
        sys.exit(1)
    except Exception as error:
        _print_err(f"Error: {error}")
        if args.debug:
            raise
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Parse the filter arguments."""

import json
from collections.abc import Mapping

from ._utility import _print_err


def _is_json_like(q):
    """Check if q is JSON like.

    Parameters
    ----------
    q : str
        Query string.

    Returns
    -------
    bool
        True if q starts with "{" and ends with "}".

    """
    return (q[0] == "{" and q[-1] == "}") or (q[0] == "[" and q[-1] == "]")


def _is_regex(q):
    """Check if q is a regular expression.

    Parameters
    ----------
    q : str
        Query string.

    Returns
    -------
    bool
        True if q starts with "/" and ends with "/".

    """
    return q.startswith("/") and q.endswith("/")


def _parse_json(q):
    """Parse a query argument as JSON.

    Parameters
    ----------
    q : json
        Query argument.

    Raises
    ------
    JSONDecodeError
        Raised if the input cannot be parsed as JSON.

    """
    try:
        return json.loads(q)
    except json.JSONDecodeError:
        _print_err(f"Failed to parse query argument. Ensure that '{q}' is valid JSON!")
        raise


CAST_MAPPING = {
    "true": True,
    "false": False,
    "null": None,
}

CAST_MAPPING_WARNING = {
    "True": "true",
    "False": "false",
    "None": "null",
    "none": "null",
}


def _cast(x):
    """Attempt to interpret x with the correct type.

    Parameters
    ----------
    x : str
        The value to cast.

    Returns
    -------
    object
        Value of x, cast from a str to an appropriate type (bool, NoneType, int, float, str).

    """
    try:
        if x in CAST_MAPPING_WARNING:
            _print_err(f"Did you mean {CAST_MAPPING_WARNING[x]}?")
        return CAST_MAPPING[x]
    except KeyError:
        try:
            return int(x)
        except ValueError:
            try:
                return float(x)
            except ValueError:
                return x


def _parse_single(key, value=None):
    """Parse simple search syntax.

    Parameters
    ----------
    key : str
        The filter key.
    value : object, optional
        The filter value. If None, the filter returns
        True if the provided key exists (Default value = None).

    Returns
    -------
    dict
        Parsed filter arguments.

    Raises
    ------
    ValueError
        If filter arguments have an invalid key.
    """
    if _is_json_like(key):
        raise ValueError(
            "Please check your filter arguments. "
            f"Using a JSON expression as a key is not allowed: '{key}'."
        )
    elif value is None or value == "!":
        return key, {"$exists": True}
    elif _is_json_like(value):
        return key, _parse_json(value)
    elif _is_regex(value):
        return key, {"$regex": value[1:-1]}
    else:
        return key, _cast(value)


def parse_simple(tokens):
    """Parse a set of string tokens into a suitable filter.

    Parameters
    ----------
    tokens : Sequence[str]
        A Sequence of strings composing key-value pairs.

    Yields
    ------
    tuple
        A single key-value pair of input tokenized filter.

    """
    for i in range(0, len(tokens), 2):
        key = tokens[i]
        if i + 1 < len(tokens):
            value = tokens[i + 1]
        else:
            value = None
        yield _parse_single(key, value)


def parse_filter_arg(args):
    """Parse a series of filter arguments into a dictionary.

    Parameters
    ----------
    args : sequence of str
        Filter arguments to parse.

    Returns
    -------
    dict
        Filter arguments.

    """
    if args is None or len(args) == 0:
        return None
    elif len(args) == 1:
        if _is_json_like(args[0]):
            return _parse_json(args[0])
        else:
            key, value = _parse_single(args[0])
            query = {key: value}
    else:
        query = dict(parse_simple(args))
    _print_err(f"Interpreted filter arguments as '{json.dumps(query)}'.")
    return query


def _add_prefix(filter):
    """Add prefix "sp." to a (possibly nested) filter."""
    for key, value in filter.items():
        if key in ("$and", "$or"):
            if isinstance(value, list) or isinstance(value, tuple):
                yield key, [dict(_add_prefix(item)) for item in value]
            else:
                raise ValueError(
                    "The argument to a logical operator must be a list or a tuple!"
                )
        elif "." in key and key.split(".", 1)[0] in ("sp", "doc"):
            yield key, value
        elif key in ("sp", "doc"):
            yield key, value
        else:
            # This line guarantees that all filter keys with no prefix
            # default to state point key searches.
            yield "sp." + key, value


def _root_keys(filter):
    for key, value in filter.items():
        if key in ("$and", "$or"):
            assert isinstance(value, (list, tuple))
            for item in value:
                for key in _root_keys(item):
                    yield key
        elif "." in key:
            yield key.split(".", 1)[0]
        else:
            yield key


def parse_filter(filter):
    """Parse a provided sequence of filters.

    Parameters
    ----------
    filter : Sequence, Mapping, or str
        A set of key, value tuples corresponding to a single filter. This
        filter may itself be a compound filter containing and/or statements. The
        filter may be provided as a sequence of tuples, a mapping-like object,
        or a string. In the last case, the string will be parsed to generate a
        valid set of filters.

    Yields
    ------
    tuple
        A key value pair to be used as a filter.

    """
    if isinstance(filter, str):
        yield from parse_simple(filter.split())
    elif isinstance(filter, Mapping):
        yield from filter.items()
    else:
        try:
            yield from filter
        except TypeError:
            # This type was not iterable.
            raise ValueError(
                f"Invalid filter type {type(filter)}. The filter must "
                "be a Sequence, Mapping, or str."
            )



import pytest

from signac._search_indexer import _SearchIndexer
from signac.errors import InvalidKeyError

n = 42
N = 100

ARITHMETIC_DOCS = {str(i): {"a": i} for i in range(N)}

ARITHMETIC_EXPRESSIONS = [
    ({"$eq": n}, 1),
    ({"$ne": n}, N - 1),
    ({"$lt": n}, n),
    ({"$gt": n}, N - n - 1),
    ({"$lte": n}, n + 1),
    ({"$gte": n}, N - n),
]


ARRAY_EXPRESSIONS = [
    ({"$in": []}, 0),
    ({"$in": [0, 1, 2]}, 3),
    ({"$in": ["a", "b", "c"]}, 0),
    ({"$nin": []}, N),
    ({"$nin": [0, 1, 2]}, N - 3),
    ({"$nin": ["a", "b", "c"]}, N),
]

LOGICAL_EXPRESSIONS = [
    ({"$and": []}, ValueError),
    ({"$and": {}}, ValueError),
    ({"$and": ""}, ValueError),
    ({"a": {"$and": []}}, KeyError),
    ({"a": {"$and": [{"b": 0}]}}, KeyError),
    ({"$and": [{"a": n}]}, 1),
    ({"$and": [{"$not": {"a": n}}]}, N - 1),
    ({"$and": [{"a": n}, {"a": n + 1}]}, 0),
    ({"$and": [{"a": n}, {"$not": {"a": n}}]}, 0),
    ({"$or": []}, ValueError),
    ({"a": {"$or": []}}, KeyError),
    ({"a": {"$or": [{"b": 0}]}}, KeyError),
    ({"$or": [{"$not": {"a": n}}]}, N - 1),
    ({"$or": [{"a": n}, {"a": n + 1}]}, 2),
    ({"$or": [{"a": n}, {"$not": {"a": n}}]}, N),
]

INVALID_SYNTAX_EXPRESSIONS = [
    ({"a": {"$$lt": N}}, KeyError),  # Bad operator expression
    ({"a": {"lt$": N}}, KeyError),  # Bad operator placement
    ({"a": {"$exists": N}}, ValueError),  # Value of $exists must be boolean
    ({"a": {"$nonexistent": N}}, KeyError),  # Unknown expression-operator
    ({"a": {"$type": N}}, ValueError),  # Invalid $type argument
    (["a"], ValueError),  # Not a valid filter (not a dict)
]


class TestSearchIndexer:
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.c = _SearchIndexer()

    def test_init(self):
        assert len(self.c) == 0

    def test_init_with_list_with_ids_sequential(self):
        docs = {str(i): {"a": i} for i in range(10)}
        self.c = _SearchIndexer(docs)
        assert len(self.c) == len(docs)
        for _id in docs:
            assert _id in self.c

    def test_init_with_list_with_ids_non_sequential(self):
        docs = {f"{i ** 3:032d}": {"a": i} for i in range(10)}
        self.c = _SearchIndexer(docs)
        assert len(self.c) == len(docs)
        for _id in docs:
            assert _id in self.c

    def test_int_float_equality(self):
        docs = {"int": {"a": 1}, "float": {"a": 1.0}}
        self.c = _SearchIndexer(docs)
        assert len(self.c.find()) == 2
        assert len(self.c.find(dict(a=1))) == 2
        assert len(self.c.find(dict(a=1.0))) == 2

    def test_copy(self):
        docs = {str(i): {"a": i} for i in range(10)}
        self.c = _SearchIndexer(docs)
        c2 = _SearchIndexer(self.c)
        assert len(self.c) == len(c2)
        for doc in c2.values():
            assert len(self.c.find(doc)) == 1

    def test_insert_and_remove(self):
        doc = {"a": 0}
        self.c["0"] = doc
        assert len(self.c) == 1
        assert self.c["0"] == doc
        assert list(self.c.find()) == ["0"]
        del self.c["0"]
        assert len(self.c) == 0
        with pytest.raises(KeyError):
            assert self.c["0"]

    def test_contains(self):
        doc = {"a": 0}
        assert "0" not in self.c
        self.c["0"] = doc
        assert "0" in self.c
        del self.c["0"]
        assert "0" not in self.c

    def test_update(self):
        docs = {str(i): {"a": i} for i in range(10)}
        self.c.update(docs)
        assert len(self.c) == len(docs)

    def test_update_collision(self):
        docs = {str(i): {"a": i} for i in range(10)}
        self.c.update(docs)
        # Update the first ten, insert the second ten
        new_docs = {str(i): {"a": i * 2} for i in range(20)}
        self.c.update(new_docs)
        assert len(self.c) == len(new_docs)
        assert self.c["0"] == {"a": 0}

    def test_index(self):
        docs = {str(i): {"a": i} for i in range(10)}
        self.c.update(docs)
        index = self.c.build_index("a")
        assert len(index) == len(self.c)
        for value, _ids in index.items():
            for _id in _ids:
                assert self.c[_id]["a"] == value

    def test_clear(self):
        assert len(self.c) == 0
        self.c["0"] = {"a": 0}
        assert len(self.c) == 1
        self.c.clear()
        assert len(self.c) == 0

    def test_iteration(self):
        assert len(self.c) == 0
        assert len(self.c.find()) == 0
        self.c["0"] = {"a": 0}
        assert len(self.c) == 1
        assert len(self.c.find()) == 1
        self.c.clear()
        docs = {str(i): {"a": i} for i in range(10)}
        self.c.update(docs)
        assert len(self.c) == len(docs)
        assert len(self.c.find()) == len(docs)
        assert {doc["a"] for doc in docs.values()} == {
            self.c[doc]["a"] for doc in self.c.find()
        }

    def test_find_id(self):
        assert len(self.c.find()) == 0
        assert len(self.c.find({"_id": "0"})) == 0
        docs = {str(i): {"a": i} for i in range(10)}
        self.c.update(docs)
        assert len(self.c.find()) == len(docs)
        assert len(self.c.find({"_id": "0"})) == 1

    def test_find_integer(self):
        assert len(self.c.find()) == 0
        assert list(self.c.find()) == []
        assert len(self.c.find({"a": 0})) == 0
        docs = {str(i): {"a": i} for i in range(10)}
        self.c.update(docs)
        assert len(self.c.find()) == len(docs)
        assert len(self.c.find({"a": 0})) == 1
        assert len(self.c.find({"a": 0.0})) == 1
        assert list(self.c.find({"a": 0})) == ["0"]
        assert len(self.c.find({"a": -1})) == 0
        assert len(self.c.find({"a.b": 0})) == 0
        assert len(self.c.find({"a": {"$type": "int"}})) == 10
        assert len(self.c.find({"a": {"$type": "float"}})) == 0
        del self.c["0"]
        assert len(self.c.find({"a": 0})) == 0

    def test_find_float(self):
        assert len(self.c.find()) == 0
        assert list(self.c.find()) == []
        assert len(self.c.find({"a": 0})) == 0
        docs = {str(i): {"a": float(i)} for i in range(10)}
        self.c.update(docs)
        assert len(self.c.find()) == len(docs)
        assert len(self.c.find({"a": 0})) == 1
        assert len(self.c.find({"a": 0.0})) == 1
        assert list(self.c.find({"a": 0.0})) == ["0"]
        assert len(self.c.find({"a": -1})) == 0
        assert len(self.c.find({"a.b": 0})) == 0
        assert len(self.c.find({"a": {"$type": "int"}})) == 0
        assert len(self.c.find({"a": {"$type": "float"}})) == 10
        del self.c["0"]
        assert len(self.c.find({"a": 0})) == 0

    def test_find_list(self):
        assert len(self.c.find()) == 0
        assert list(self.c.find()) == []
        assert len(self.c.find({"a": []})) == 0
        docs = {"0": {"a": []}}
        self.c.update(docs)
        assert len(self.c.find()) == 1
        assert len(self.c.find({"a": []})) == 1
        for i, v in enumerate((None, 1, "1", {"b": 1}), 1):
            docs = {str(i): {"a": [v]}}
            self.c.update(docs)
            assert len(self.c.find({"a": [v]})) == 1

    def test_find_int_float(self):
        docs = {"int": {"a": 1}, "float": {"a": 1.0}}
        self.c.update(docs)
        assert len(self.c.find({"a": {"$type": "float"}})) == 1
        assert len(self.c.find({"a": {"$type": "int"}})) == 1
        assert list(self.c.find({"a": {"$type": "float"}})) == ["float"]
        assert list(self.c.find({"a": {"$type": "int"}})) == ["int"]

    def test_docs_with_dots(self):
        # Dots are not allowed in keys of the document. Here, we explicitly
        # test to ensure that an error is raised if invalid keys are found.
        self.c["0"] = {"a.b": 0}

        # These searches will not trigger an error:
        self.c.find()
        self.c.find({"a": 0})

        # This search triggers an error:
        with pytest.raises(InvalidKeyError):
            self.c.find({"a.b": 0})

    def test_find_types(self):
        # Note: All of the iterables will be normalized to lists!
        t = [1, 1.0, "1", [1], tuple([1])]
        for i, t in enumerate(t):
            self.c.clear()
            doc = self.c[str(i)] = {"a": t}
            assert list(self.c.find(doc)) == [str(i)]

    def test_find_nested(self):
        docs = {str(i): {"a": {"b": i}} for i in range(10)}
        self.c.update(docs)
        assert len(self.c.find()) == len(docs)
        assert len(self.c.find({"a.b": 0})) == 1
        assert len(self.c.find({"a": {"b": 0}})) == 1
        assert list(self.c.find({"a.b": 0})) == ["0"]
        del self.c["0"]
        assert len(self.c.find({"a.b": 0})) == 0
        assert len(self.c.find({"a": {"b": 0}})) == 0

    def test_nested_lists(self):
        docs = {str(i): {"a": [[[i]]]} for i in range(10)}
        self.c.update(docs)
        assert len(self.c.find()) == len(docs)
        assert len(self.c.find({"a": [[[-1]]]})) == 0
        assert len(self.c.find({"a": [[[0]]]})) == 1

    def test_find_exists_operator(self):
        assert len(self.c) == 0
        data = {
            "a": True,
            "b": "b",
            "c": 0,
            "d": 0.1,
            "e": {"a": 0},
            "f": {"a": "b"},
            "g": [0, "a", True],
        }

        # Test without data
        for key in data:
            assert len(self.c.find({key: {"$exists": False}})) == len(self.c)
            assert len(self.c.find({key: {"$exists": True}})) == 0
            assert len(self.c.find({f"{key}.$exists": False})) == len(self.c)
            assert len(self.c.find({f"{key}.$exists": True})) == 0

        # Test for nested cases
        assert len(self.c.find({"e.a.$exists": True})) == 0
        assert len(self.c.find({"e.a.$exists": False})) == 0
        assert len(self.c.find({"e.a": {"$exists": True}})) == 0
        assert len(self.c.find({"e.a": {"$exists": False}})) == 0
        assert len(self.c.find({"f.a.$exists": True})) == 0
        assert len(self.c.find({"f.a.$exists": False})) == 0
        assert len(self.c.find({"f.a": {"$exists": True}})) == 0
        assert len(self.c.find({"f.a": {"$exists": False}})) == 0

        # Test with data
        for i, (key, value) in enumerate(data.items()):
            self.c[str(i)] = {key: value}

        # Heterogeneous nesting
        self.c[str(len(self.c))] = {"e": -1}

        for key in data:
            n = 2 if key == "e" else 1
            assert len(self.c.find({key: {"$exists": False}})) == len(self.c) - n
            assert len(self.c.find({key: {"$exists": True}})) == n
            assert len(self.c.find({f"{key}.$exists": False})) == len(self.c) - n
            assert len(self.c.find({f"{key}.$exists": True})) == n

        # Test for nested cases
        assert len(self.c.find({"e.$exists": True})) == 2
        assert len(self.c.find({"e.a.$exists": True})) == 1
        assert len(self.c.find({"e.a.$exists": False})) == len(self.c) - 1
        assert len(self.c.find({"e.a": {"$exists": True}})) == 1
        assert len(self.c.find({"e.a": {"$exists": False}})) == len(self.c) - 1
        assert len(self.c.find({"f.a.$exists": True})) == 1
        assert len(self.c.find({"f.a.$exists": False})) == len(self.c) - 1
        assert len(self.c.find({"f.a": {"$exists": True}})) == 1
        assert len(self.c.find({"f.a": {"$exists": False}})) == len(self.c) - 1

    def test_find_arithmetic_operators(self):
        assert len(self.c) == 0
        for expr, n in ARITHMETIC_EXPRESSIONS:
            assert len(self.c.find({"a": expr})) == 0
        self.c.update(ARITHMETIC_DOCS)
        assert len(self.c) == len(ARITHMETIC_DOCS)
        for expr, n in ARITHMETIC_EXPRESSIONS:
            assert len(self.c.find({"a": expr})) == n

    def test_find_near(self):
        assert len(self.c) == 0
        # find 0 items in empty collection
        assert len(self.c.find({"a": {"$near": [10]}})) == 0
        assert len(self.c.find({"a": {"$near": [10, 100]}})) == 0
        assert len(self.c.find({"a": {"$near": [10, 100, 100]}})) == 0
        assert len(self.c.find({"a": {"$near": (10)}})) == 0
        assert len(self.c.find({"a": {"$near": (10, 100)}})) == 0
        assert len(self.c.find({"a": {"$near": (10, 100, 100)}})) == 0
        self.c.update(ARITHMETIC_DOCS)
        assert len(self.c) == len(ARITHMETIC_DOCS)
        # test known cases with lists and tuples
        assert len(self.c.find({"a": {"$near": [10]}})) == 1
        assert len(self.c.find({"a": {"$near": (10)}})) == 1
        assert len(self.c.find({"a": {"$near": [10]}})) == len(
            self.c.find({"a": {"$near": (10)}})
        )
        assert len(self.c.find({"a": {"$near": [10]}})) == len(
            self.c.find({"a": {"$near": 10}})
        )
        assert len(self.c.find({"a": {"$near": [10, 0.5]}})) == 16
        assert len(self.c.find({"a": {"$near": (10, 0.5)}})) == 16
        assert len(self.c.find({"a": {"$near": [10, 0.5, 0.0]}})) == 16
        assert len(self.c.find({"a": {"$near": (10, 0.5, 0.0)}})) == 16
        # increasing abs_tol should increase # of jobs found
        assert len(self.c.find({"a": {"$near": [10, 0.5, 11]}})) > len(
            self.c.find({"a": {"$near": [10, 0.5]}})
        )
        assert len(self.c.find({"a": {"$near": [10.5, 0.005]}})) == 0
        assert len(self.c.find({"a": {"$near": (10.5, 0.005)}})) == 0
        # test with lists that are too long
        with pytest.raises(ValueError):
            self.c.find({"a": {"$near": [10, 0.5, 1, 1]}})
        with pytest.raises(ValueError):
            self.c.find({"a": {"$near": [10, 0.5, 1, 1, 5]}})
        with pytest.raises(ValueError):
            self.c.find({"a": {"$near": (10, 0.5, 1, 1)}})
        with pytest.raises(ValueError):
            self.c.find({"a": {"$near": (10, 0.5, 1, 1, 5)}})

    def test_find_array_operators(self):
        assert len(self.c) == 0
        for expr, n in ARRAY_EXPRESSIONS:
            assert len(self.c.find({"a": expr})) == 0
        self.c.update(ARITHMETIC_DOCS)
        assert len(self.c) == len(ARITHMETIC_DOCS)
        for expr, n in ARRAY_EXPRESSIONS:
            assert len(self.c.find({"a": expr})) == n

    def test_find_regular_expression(self):
        assert len(self.c) == 0
        assert len(self.c.find({"a": {"$regex": "foo"}})) == 0
        assert len(self.c.find({"a": {"$regex": "hello"}})) == 0
        docs = {"0": {"a": "hello world"}}
        self.c.update(docs)
        assert len(self.c.find({"a": {"$regex": "foo"}})) == 0
        assert len(self.c.find({"a": {"$regex": "hello"}})) == 1
        assert len(self.c.find({"a": {"$regex": "hello world"}})) == 1

    def test_find_type_expression(self):
        assert len(self.c) == 0
        types = [
            (1, "int"),
            (1.0, "float"),
            ("1.0", "str"),
            (True, "bool"),
            (None, "null"),
        ]
        for v, t in types:
            assert len(self.c.find({"a": {"$type": t}})) == 0
        for i, (v, t) in enumerate(types):
            self.c[str(i)] = {str(i): v}
        assert len(self.c) == len(types)
        for i, (v, t) in enumerate(types):
            assert len(self.c.find({str(i): {"$type": t}})) == 1

    def test_find_type_integer_values_identical_keys(self):
        docs = {"int": {"a": 1}, "float": {"a": 1.0}}
        self.c = _SearchIndexer(docs)
        assert len(self.c.find({"a": {"$type": "int"}})) == 1
        assert len(self.c.find({"a": {"$type": "float"}})) == 1

    def test_find_where_expression(self):
        assert len(self.c) == 0
        assert len(self.c.find({"a": {"$where": "lambda x: x < 42"}})) == 0
        self.c.update(ARITHMETIC_DOCS)
        assert len(self.c.find({"a": {"$where": "lambda x: x < 42"}})) == 42

    def test_find_logical_operators(self):
        assert len(self.c) == 0
        for expr, expectation in LOGICAL_EXPRESSIONS:
            if not isinstance(expectation, int):
                with pytest.raises(expectation):
                    self.c.find(expr)
            else:
                assert len(self.c.find(expr)) == 0
        self.c.update(ARITHMETIC_DOCS)
        assert len(self.c) == len(ARITHMETIC_DOCS)
        for expr, expectation in LOGICAL_EXPRESSIONS:
            if not isinstance(expectation, int):
                with pytest.raises(expectation):
                    self.c.find(expr)
            else:
                assert len(self.c.find(expr)) == expectation
                assert len(self.c.find({"$not": expr})) == N - expectation
                assert len(self.c.find({"$not": {"$not": expr}})) == expectation

    def test_find_invalid_syntax(self):
        self.c.update(ARITHMETIC_DOCS)
        assert len(self.c) == len(ARITHMETIC_DOCS)
        for expr, expectation in INVALID_SYNTAX_EXPRESSIONS:
            with pytest.raises(expectation):
                self.c.find(expr)



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Linked view classes."""

import errno
import logging
import os
import sys
from itertools import chain

from ._utility import _mkdir_p

logger = logging.getLogger(__name__)


def create_linked_view(project, prefix=None, job_ids=None, path=None):
    """Create or update a persistent linked view of the selected data space.

    Parameters
    ----------
    project : signac.Project
        Project handle.
    prefix : str, optional
        The path where the linked view will be created or updated (Default value = None).
    job_ids : iterable, optional
        If None (the default), create the view for the complete data space,
        otherwise only for this iterable of job ids.
    path : str or callable, optional
        The path (function) used to structure the linked data space (Default value = None).

    Returns
    -------
    dict
        A dictionary that maps the source directory paths to the linked
        directory paths.

    Raises
    ------
    OSError
        Linked views cannot be created on Windows because
        symbolic links are not supported by the platform.
    RuntimeError
        When state points contain one of ``[os.sep, " ", "*"]``.

    """
    from .import_export import _check_directory_structure_validity, _make_path_function

    # Windows does not support the creation of symbolic links.
    if sys.platform == "win32":
        raise OSError(
            "signac cannot create linked views on Windows, because "
            "symbolic links are not supported by the platform."
        )

    if prefix is None:
        prefix = "view"

    if job_ids is None:
        jobs = list(project)
    else:
        jobs = list(project.open_job(id=job_id) for job_id in job_ids)

    key_list = [k for job in jobs for k in job.statepoint().keys()]
    value_list = [v for job in jobs for v in job.statepoint().values()]
    item_list = key_list + value_list
    bad_chars = [os.sep, " ", "*"]
    bad_items = [
        item
        for item in item_list
        for char in bad_chars
        if isinstance(item, str) and char in item
    ]

    if any(bad_items):
        err_msg = " ".join(
            [
                f"In order to use view, state points should not contain {bad_chars}:",
                *bad_items,
            ]
        )
        raise RuntimeError(err_msg)

    path_function = _make_path_function(jobs, path)

    links = {}
    for job in jobs:
        paths = os.path.join(path_function(job), "job")
        links[paths] = job.path
    if not links:  # data space contains less than two elements
        for job in project.find_jobs():
            links["./job"] = job.path
        assert len(links) < 2
    _check_directory_structure_validity(links.keys())
    _update_view(prefix, links)

    return links


def _update_view(prefix, links, leaf="job"):
    """Update an existing linked view hierarchy in place.

    Parameters
    ----------
    prefix : str
        The path where the linked view will be created or updated (Default value = None).
    links : dict
        Linked view .
    leaf : str
        The name of the leaf directories in the view
        directory tree (Default value = 'job').

    """
    obsolete, to_update, new = _analyze_view(prefix, links)
    num_ops = len(obsolete) + 2 * len(to_update) + len(new)
    if num_ops:
        logger.info(f"Generating current view in '{prefix}' ({num_ops} operations)...")
    else:
        logger.info(f"View in '{prefix}' is up to date.")
        return
    logger.debug(f"Removing {len(obsolete)} obsolete links.")
    for path in obsolete:
        p = os.path.join(prefix, path)
        try:
            os.unlink(p)
        except OSError:
            os.rmdir(p)
    logger.debug(
        "Creating {} new and updating {} existing links.".format(
            len(new), len(to_update)
        )
    )
    for path in to_update:
        os.unlink(os.path.join(prefix, path))
    for path in chain(new, to_update):
        dst = os.path.join(prefix, path)
        src = os.path.relpath(links[path], os.path.split(dst)[0])
        _make_link(src, dst)


def _analyze_view(prefix, links, leaf="job"):
    """Analyze an existing view to prepare for update.

    Parameters
    ----------
    prefix : str
        The path where the linked view will be created or updated (Default value = None).
    links : dict
        Linked view.
    leaf : str
        The name of the leaf directories in the view
        directory tree (Default value = 'job').

    Returns
    -------
    tuple
        tuple that contains: (list of outdated links, list of links to update, set of new links).

    """
    logger.info(f"Analyzing view prefix '{prefix}'...")
    existing_paths = {os.path.join(p, leaf) for p in _find_all_links(prefix, leaf)}
    existing_tree = _build_tree(existing_paths)
    for path in links:
        _color_path(existing_tree, path.split(os.sep))
    obsolete = []
    dead_branches = _find_dead_branches(existing_tree)
    for branch in reversed(sorted(dead_branches, key=len)):
        if branch:
            obsolete.append(os.path.join(*(n.name for n in branch)))
    if "." in obsolete:
        obsolete.remove(".")
    keep_or_update = existing_paths.intersection(links.keys())
    new = set(links.keys()).difference(keep_or_update)
    to_update = [
        p
        for p in keep_or_update
        if os.path.realpath(os.path.join(prefix, p)) != links[p]
    ]
    return obsolete, to_update, new


def _make_link(src, dst):
    """Create a symbolic link and all directories leading to it.

    Parameters
    ----------
    src : str
        Name of directory/file to create a symbolic link.
    dst : str
        Destination symbolic link directory/file name.

    """
    _mkdir_p(os.path.dirname(dst))
    try:
        os.symlink(src, dst, target_is_directory=True)
    except OSError as error:
        if error.errno == errno.EEXIST:
            if os.path.realpath(src) == os.path.realpath(dst):
                return
        raise


def _find_all_links(root, leaf="job"):
    """Find all symbolic links under root.

    Parameters
    ----------
    root : str
        Project root directory.
    leaf : str
        The name of the leaf directories in the view
        directory tree (Default value = 'job').

    Yields
    ------
    str
        Relative path to root from ``leaf``.

    """
    for dirpath, dirnames, filenames in os.walk(root):
        for dirname in dirnames:
            if dirname == leaf:
                yield os.path.relpath(dirpath, root)
                break
        for filename in filenames:
            if filename == leaf:
                yield os.path.relpath(dirpath, root)
                break


class _Node:
    """Generic graph-node class."""

    def __init__(self, name=None, value=None):
        self.name = name
        self.value = value
        self.children = {}

    def get_child(self, name):
        """Get child node corresponding to the name passed.

        Parameters
        ----------
        name : str
            Name of child node to get.

        Returns
        -------
        :class:`~signac.linked_view._Node`
            The requested child node.

        """
        return self.children.setdefault(name, type(self)(name))

    def __str__(self):
        return f"_Node({self.name}, {self.value})"

    __repr__ = __str__


def _build_tree(paths):
    """Build a graph structure for paths.

    Parameters
    ----------
    paths : list
        A list of paths to views that already exist.

    Returns
    -------
    :class:`~signac.linked_view._Node`
        Graph structure for path.

    """
    root = _Node()
    for path in paths:
        node = root
        for p in path.split(os.sep):
            node = node.get_child(p)
    return root


def _color_path(root, path):
    """Color the path from root by setting value to True.

    Parameters
    ----------
    root : :class:`~signac.linked_view._Node`
        Root node.
    path : list
        The name of the directory/file to color (set value to True).

    """
    root.value = True
    for name in path:
        root = root.get_child(name)
        root.value = True


def _find_dead_branches(root, branch=None):
    """Find all branches considered dead (not-colored).

    Parameters
    ----------
    root : :class:`~signac.linked_view._Node`
        Root node.
    branch : list, optional
        The current list of branches that have been collected,
        used in recursive calls to build up the branches starting
        at the root (Default value = None).

    Yields
    ------
    list
        Branches that are considered as dead (not-colored).

    """
    if branch is None:
        branch = []
    else:
        branch = list(branch) + [root]
    if root.children:
        for child in root.children.values():
            yield from _find_dead_branches(child, branch)
    if not root.value:
        yield branch



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import pytest
from test_project import TestProjectBase

try:
    import pandas  # noqa

    PANDAS = True
except ImportError:
    PANDAS = False


@pytest.mark.skipif(not PANDAS, reason="test requires the pandas package")
class TestPandasIntegration(TestProjectBase):
    def test_to_dataframe(self):
        for i in range(10):
            job = self.project.open_job(dict(a=i))
            job.doc.b = float(i)
        df = self.project.to_dataframe()
        assert len(df) == len(self.project)
        assert "sp.a" in df.columns
        assert "doc.b" in df.columns

    def test_jobs_iterator_to_dataframe(self):
        for i in range(10):
            job = self.project.open_job(dict(a=i))
            job.doc.b = float(i)
        df = self.project.find_jobs().to_dataframe()
        assert len(df) == len(self.project)
        assert "sp.a" in df.columns
        assert "doc.b" in df.columns
        jobs = self.project.find_jobs(dict({"a.$lte": 5}))
        df = jobs.to_dataframe()
        assert len(df) == len(jobs)
        assert "sp.a" in df.columns
        assert "doc.b" in df.columns

    def test_prefixes(self):
        for i in range(10):
            job = self.project.open_job(dict(a=i))
            job.doc.b = float(i)
        df = self.project.to_dataframe(sp_prefix="", doc_prefix="")
        assert len(df) == len(self.project)
        assert "a" in df.columns
        assert "b" in df.columns

    def test_usecols(self):
        for i in range(10):
            job = self.project.open_job(dict(a=i, b=i * 2))
            job.doc.c = float(i)
            job.doc.d = float(i * 3)

        # Including no keys should return an empty DataFrame
        df = self.project.to_dataframe(usecols=[])
        assert len(df.columns) == 0
        assert len(df) == 0

        # Excluding all keys should return an empty DataFrame
        def usecols(column):
            return column not in ["sp.a", "sp.b", "doc.c", "doc.d"]

        df = self.project.to_dataframe(usecols=usecols)
        assert len(df.columns) == 0
        assert len(df) == 0

        # Include one state point column
        df = self.project.to_dataframe(usecols=["sp.a"])
        assert "sp.a" in df.columns
        assert len(df.columns) == 1
        assert len(df) == len(self.project)

        # Exclude one state point column
        def usecols(column):
            return column != "sp.b"

        df = self.project.to_dataframe(usecols=usecols)
        assert "sp.a" in df.columns
        assert "sp.b" not in df.columns
        assert "doc.c" in df.columns
        assert "doc.d" in df.columns
        assert len(df.columns) == 3
        assert len(df) == len(self.project)

        # Include one document column
        df = self.project.to_dataframe(usecols=["doc.c"])
        assert "doc.c" in df.columns
        assert len(df.columns) == 1
        assert len(df) == len(self.project)

        # Exclude one document column
        def usecols(column):
            return column != "doc.d"

        df = self.project.to_dataframe(usecols=usecols)
        assert "sp.a" in df.columns
        assert "sp.b" in df.columns
        assert "doc.c" in df.columns
        assert "doc.d" not in df.columns
        assert len(df.columns) == 3
        assert len(df) == len(self.project)

    def test_flatten(self):
        for i in range(10):
            job = self.project.open_job(dict(a=dict(b=i * 2, c=i * 3), d=i))
            job.doc.e = dict(f=float(i))

        # Including no keys should return an empty DataFrame
        df = self.project.to_dataframe(usecols=[], flatten=True)
        assert len(df.columns) == 0
        assert len(df) == 0

        # Include one flattened state point column
        df = self.project.to_dataframe(usecols=["sp.a.b"], flatten=True)
        assert "sp.a.b" in df.columns
        assert len(df.columns) == 1
        assert len(df) == len(self.project)

        # Include one flattened document column
        df = self.project.to_dataframe(usecols=["doc.e.f"], flatten=True)
        assert "doc.e.f" in df.columns
        assert len(df.columns) == 1
        assert len(df) == len(self.project)



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from tempfile import TemporaryDirectory

import pytest

import signac


class TestDiffBase:
    project_class = signac.Project

    @pytest.fixture(autouse=True)
    def setUp(self, request):
        self._tmp_dir = TemporaryDirectory(prefix="signac_")
        request.addfinalizer(self._tmp_dir.cleanup)
        self.project = self.project_class.init_project(path=self._tmp_dir.name)


class TestDiff(TestDiffBase):
    def test_two_jobs(self):
        job1 = self.project.open_job({"a": 0, "b": 1})
        job2 = self.project.open_job({"a": 0})
        expected = {str(job1.id): {"b": 1}, str(job2.id): {}}
        result = signac.diff_jobs(job1, job2)
        assert expected == result, f"{result} is not {expected}"

    def test_one_job(self):
        job1 = self.project.open_job({"a": 0})
        expected = {str(job1.id): {}}
        result = signac.diff_jobs(job1)
        assert expected == result, f"{result} is not {expected}"

    def test_no_jobs(self):
        assert signac.diff_jobs() == {}

    def test_nested(self):
        job1 = self.project.open_job({"a": 0, "b": {"c": True, "d": 11}})
        job2 = self.project.open_job({"a": 0, "b": {"c": True, "d": 4}})
        expected = {str(job1.id): {"b": {"d": 11}}, str(job2.id): {"b": {"d": 4}}}
        result = signac.diff_jobs(job1, job2)
        assert expected == result, f"{result} is not {expected}"

    def test_same_job(self):
        job1 = self.project.open_job({"a": 0, "b": 1})
        expected = {str(job1.id): {}}
        result = signac.diff_jobs(job1, job1)
        assert expected == result, f"{result} is not {expected}"



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import json
import os
import shutil
import subprocess
import sys
from tempfile import TemporaryDirectory

import pytest
from test_project import _initialize_v1_project

import signac
from signac._config import USER_CONFIG_FN, _Config, _load_config, _read_config_file

# Skip linked view tests on Windows
WINDOWS = sys.platform == "win32"


class DummyFile:
    "We redirect sys stdout into this file during console tests."

    def __init__(self):
        self._x = ""

    def write(self, x):
        self._x += x

    def flush(self):
        pass

    def read(self):
        x = self._x
        self._x = ""
        return x


class ExitCodeError(RuntimeError):
    pass


class TestBasicShell:
    @pytest.fixture(autouse=True)
    def setUp(self, request):
        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is None:
            pythonpath = [os.getcwd()]
        else:
            pythonpath = [os.getcwd()] + pythonpath.split(":")
        os.environ["PYTHONPATH"] = ":".join(pythonpath)
        self.tmpdir = TemporaryDirectory(prefix="signac_")
        request.addfinalizer(self.tmpdir.cleanup)
        self.cwd = os.getcwd()
        os.chdir(self.tmpdir.name)
        request.addfinalizer(self.return_to_cwd)

    def return_to_cwd(self):
        os.chdir(self.cwd)

    def call(self, command, input=None, shell=False, error=False, raise_error=True):
        p = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=shell,
        )
        if input:
            p.stdin.write(input.encode())
        out, err = p.communicate()
        if p.returncode != 0 and raise_error:
            raise ExitCodeError(f"STDOUT='{out}' STDERR='{err}'")
        return err.decode() if error else out.decode()

    def test_print_usage(self):
        with pytest.raises(ExitCodeError):
            self.call("python -m signac".split())

        out = self.call("python -m signac".split(), raise_error=False)
        assert "usage:" in out

    def test_version(self):
        out = self.call("python -m signac --version".split())
        assert f"signac {signac.__version__}" in out

    def test_help(self):
        out = self.call("python -m signac --help".split())
        assert "positional arguments:" in out
        # Help string changed in 3.10.
        py310_or_greater = sys.version_info >= (3, 10)
        assert ("options:" if py310_or_greater else "optional arguments:") in out

    def test_init_project(self):
        self.call("python -m signac init".split())
        assert signac.get_project().path == os.getcwd()

    def test_job_with_argument(self):
        self.call("python -m signac init".split())
        assert (
            self.call(["python", "-m", "signac", "job", '{"a": 0}']).strip()
            == "9bfd29df07674bc4aa960cf661b5acd2"
        )

    def test_job_with_argument_workspace(self):
        self.call("python -m signac init".split())
        wd_path = os.path.join(
            self.tmpdir.name, "workspace", "9bfd29df07674bc4aa960cf661b5acd2"
        )
        assert os.path.realpath(
            self.call(["python", "-m", "signac", "job", "--path", '{"a": 0}']).strip()
        ) == os.path.realpath(wd_path)

    def test_job_with_argument_create_workspace(self):
        self.call("python -m signac init".split())
        wd_path = os.path.join(
            self.tmpdir.name, "workspace", "9bfd29df07674bc4aa960cf661b5acd2"
        )
        assert not os.path.isdir(wd_path)
        self.call(["python", "-m", "signac", "job", "--create", '{"a": 0}'])
        assert os.path.isdir(wd_path)

    def test_statepoint(self):
        self.call("python -m signac init".split())
        self.call(["python", "-m", "signac", "job", "--create", '{"a": 0}'])
        project = signac.Project()
        assert len(project) == 1
        job = project.open_job({"a": 0})
        sp = self.call(f"python -m signac statepoint {job.id}".split())
        assert project.open_job(json.loads(sp)) == job
        assert len(project) == 1
        sp = self.call("python -m signac statepoint".split())
        assert project.open_job(json.loads(sp)) == job
        assert len(project) == 1
        sp = self.call("python -m signac statepoint --pretty".split())
        assert "{'a': 0}" in sp
        assert len(project) == 1

    def test_document(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        job_a = project.open_job({"a": 0})
        job_a.init()
        assert len(project) == 1
        job_a.document["data"] = 4
        doc = json.loads(self.call("python -m signac document".split()))
        assert "data" in doc
        assert doc["data"] == 4
        doc = json.loads(self.call(f"python -m signac document {job_a.id}".split()))
        assert "data" in doc
        assert doc["data"] == 4
        out = self.call("python -m signac document --pretty".split())
        for key, value in doc.items():
            assert str(key) in out
            assert str(value) in out

    @pytest.mark.skipif(WINDOWS, reason="Symbolic links are unsupported on Windows.")
    def test_view_single(self):
        """Check whether command line views work for single job workspaces."""
        self.call("python -m signac init".split())
        project = signac.Project()
        sps = [{"a": i} for i in range(1)]
        for sp in sps:
            project.open_job(sp).init()
        os.mkdir("view")
        self.call("python -m signac view".split())
        for sp in sps:
            assert os.path.isdir("view/job")
            assert os.path.realpath("view/job") == os.path.realpath(
                project.open_job(sp).path
            )

    @pytest.mark.skipif(WINDOWS, reason="Symbolic links are unsupported on Windows.")
    def test_view(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        sps = [{"a": i} for i in range(3)]
        for sp in sps:
            project.open_job(sp).init()
        os.mkdir("view")
        self.call("python -m signac view".split())
        for sp in sps:
            assert os.path.isdir("view/a/{}".format(sp["a"]))
            assert os.path.isdir("view/a/{}/job".format(sp["a"]))
            assert os.path.realpath(
                "view/a/{}/job".format(sp["a"])
            ) == os.path.realpath(project.open_job(sp).path)

    @pytest.mark.skipif(WINDOWS, reason="Symbolic links are unsupported on Windows.")
    def test_view_prefix(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        sps = [{"a": i} for i in range(3)]
        for sp in sps:
            project.open_job(sp).init()
        os.mkdir("view")
        self.call("python -m signac view --prefix view/test_dir".split())
        for sp in sps:
            assert os.path.isdir("view/test_dir/a/{}".format(sp["a"]))
            assert os.path.isdir("view/test_dir/a/{}/job".format(sp["a"]))
            assert os.path.realpath(
                "view/test_dir/a/{}/job".format(sp["a"])
            ) == os.path.realpath(project.open_job(sp).path)

    @pytest.mark.skipif(WINDOWS, reason="Symbolic links are unsupported on Windows.")
    def test_view_incomplete_path_spec(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        sps = [{"a": i} for i in range(3)]
        for sp in sps:
            project.open_job(sp).init()
        os.mkdir("view")

        # An error should be raised if the user-provided path function
        # doesn't make a 1-1 mapping.
        err = self.call(
            "python -m signac view non_unique".split(),
            error=True,
            raise_error=False,
        )
        assert "duplicate paths" in err

    def test_find(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        sps = [{"a": i} for i in range(3)]
        sps.append({"a": [0, 1, 0]})
        for sp in sps:
            project.open_job(sp).init()
        out = self.call("python -m signac find".split())
        job_ids = out.split(os.linesep)[:-1]
        assert set(job_ids) == {job.id for job in project.find_jobs()}
        assert (
            self.call("python -m signac find".split() + ['{"a": 0}']).strip()
            == next(iter(project.find_jobs({"a": 0}))).id
        )

        job = project.open_job({"a": 0})
        out = self.call("python -m signac find a 0 --sp".split()).strip()
        assert out.strip().split(os.linesep) == [str(job.id), str(job.statepoint())]
        out = self.call("python -m signac find a 0 --sp a".split()).strip()
        assert out.strip().split(os.linesep) == [str(job.id), str(job.statepoint())]
        out = self.call("python -m signac find a 0 --sp b".split()).strip()
        assert out.strip().split(os.linesep) == [str(job.id), "{}"]

        job.document["a"] = 2
        out = self.call("python -m signac find a 0 --doc".split()).strip()
        assert out.strip().split(os.linesep) == [str(job.id), str(job.document)]
        out = self.call("python -m signac find a 0 --doc a".split()).strip()
        assert out.strip().split(os.linesep) == [str(job.id), str(job.document)]
        out = self.call("python -m signac find a 0 --doc b".split()).strip()
        assert out.strip().split(os.linesep) == [str(job.id), "{}"]
        out = self.call("python -m signac find a 0 --show --one-line".split()).strip()
        assert str(job.id) in out
        assert '{"a": 0}' in out
        assert '{"a": 2}' in out

        job = project.open_job({"a": [0, 1, 0]})
        msg = [*"python -m signac find a".split(), "[0, 1, 0]", "--sp"]
        out = self.call(msg).strip()
        assert out.strip().split(os.linesep) == [str(job.id), str(job.statepoint)]

        # Test the doc_filter
        for job in project.find_jobs():
            if job.statepoint()["a"] == [0, 1, 0]:
                continue
            job.document["a"] = job.statepoint()["a"]
            job.document["b"] = job.statepoint()["a"] + 1

        for i in range(3):
            assert (
                self.call(
                    "python -m signac find ".split() + ['{"doc.a": ' + str(i) + "}"]
                ).strip()
                == [job.id for job in project.find_jobs({"doc.a": i})][0]
            )

        for i in range(1, 4):
            assert (
                self.call(
                    "python -m signac find ".split() + ['{"doc.b": ' + str(i) + "}"]
                ).strip()
                == [job.id for job in project.find_jobs({"doc.b": i})][0]
            )

    def test_diff(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        job_a = project.open_job({"a": 0, "b": 1})
        job_a.init()
        job_b = project.open_job({"a": 0, "b": 0})
        job_b.init()
        out = self.call(f"python -m signac diff {job_a.id} {job_b.id}".split())
        expected = [str(job_a.id), "{'b': 1}", str(job_b.id), "{'b': 0}"]
        outputs = out.strip().split(os.linesep)
        assert set(expected) == set(outputs)

    def test_clone(self):
        self.call("python -m signac init".split())
        project_a = signac.Project()
        project_b = signac.init_project(path=os.path.join(self.tmpdir.name, "b"))
        job = project_a.open_job({"a": 0})
        job.init()
        assert len(project_a) == 1
        assert len(project_b) == 0

        self.call(
            "python -m signac clone {} {}".format(
                os.path.join(self.tmpdir.name, "b"), job.id
            ).split()
        )
        assert len(project_a) == 1
        assert job in project_a
        assert len(project_b) == 1
        assert job in project_b

        # cloning a job that exist at both source and destination
        err = self.call(
            "python -m signac clone {} {}".format(
                os.path.join(self.tmpdir.name, "b"), job.id
            ).split(),
            error=True,
        )
        assert "Destination already exists" in err
        assert len(project_a) == 1
        assert job in project_a
        assert len(project_b) == 1
        assert job in project_b

        # checking for id that does not exit at source
        with pytest.raises(ExitCodeError):
            self.call(
                "python -m signac clone {} 9bfd29df07674bc5".format(
                    os.path.join(self.tmpdir.name, "b")
                ).split()
            )
        assert len(project_a) == 1
        assert len(project_b) == 1

    def test_move(self):
        self.call("python -m signac init".split())
        project_a = signac.Project()
        project_b = signac.init_project(path=os.path.join(self.tmpdir.name, "b"))
        job = project_a.open_job({"a": 0})
        job.init()
        assert len(project_a) == 1
        assert len(project_b) == 0

        self.call(
            "python -m signac move {} {}".format(
                os.path.join(self.tmpdir.name, "b"), job.id
            ).split()
        )
        assert len(project_a) == 0
        assert job not in project_a
        assert len(project_b) == 1
        assert job in project_b

        # moving a job that already exists at destination
        project_a.open_job({"a": 0}).init()
        err = self.call(
            "python -m signac move {} {}".format(
                os.path.join(self.tmpdir.name, "b"), job.id
            ).split(),
            error=True,
        )
        assert "Destination already exists" in err
        assert len(project_a) == 1
        assert job in project_a
        assert len(project_b) == 1
        assert job in project_b

        # moving a job that does not exits
        with pytest.raises(ExitCodeError):
            self.call(
                "python -m signac move {} 9bfd29df07674bc5".format(
                    os.path.join(self.tmpdir.name, "b")
                ).split()
            )
        assert len(project_a) == 1
        assert len(project_b) == 1

    def test_remove(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        sps = [{"a": i} for i in range(3)]
        for sp in sps:
            project.open_job(sp).init()
        job_to_remove = project.open_job({"a": 1})
        job_to_remove.doc.a = 0
        assert job_to_remove in project
        assert job_to_remove.doc.a == 0
        assert len(job_to_remove.doc) == 1
        self.call(f"python -m signac rm --clear {job_to_remove.id}".split())
        assert job_to_remove in project
        assert len(job_to_remove.doc) == 0
        self.call(f"python -m signac -v rm {job_to_remove.id}".split())
        assert job_to_remove not in project

        # removing job that does not exist at source
        with pytest.raises(ExitCodeError):
            self.call(f"python -m signac -v rm {job_to_remove.id}".split())
        assert job_to_remove not in project

    def test_schema(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        for i in range(10):
            project.open_job(
                {
                    "a": i,
                    "b": {"b2": i},
                    "c": [i if i % 2 else None, 0, 0],
                    "d": [[i, 0, 0]],
                    "e": {"e2": [i, 0, 0]} if i % 2 else 0,  # heterogeneous!
                    "f": {"f2": [[i, 0, 0]]},
                }
            ).init()

        s = project.detect_schema()
        out = self.call("python -m signac schema".split())
        assert s.format() == out.strip().replace(os.linesep, "\n")

    def test_sync(self):
        project_b = signac.init_project(path=os.path.join(self.tmpdir.name, "b"))
        self.call("python -m signac init".split())
        project_a = signac.Project()
        for i in range(4):
            project_a.open_job({"a": i}).init()
            project_b.open_job({"a": i}).init()
        job_src = project_b.open_job({"a": 0})
        job_dst = project_a.open_job({"a": 0})
        with job_src:
            with open("test", "w") as file:
                file.write("x")
        assert len(project_a) == 4
        assert len(project_b) == 4
        project_b.document["a"] = 0
        project_a.document["b"] = 0
        out = self.call(
            "python -m signac sync {} {} --stats --human-readable".format(
                os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
            ).split()
        )
        assert "Number of files transferred: 1" in out
        assert len(project_a) == 4
        assert len(project_b) == 4
        assert "a" in project_a.document
        assert "a" in project_b.document
        assert "b" in project_a.document
        assert "b" not in project_b.document
        with job_dst:
            with open("test") as file:
                assert "x" == file.read()
        # invalid cases
        with pytest.raises(ExitCodeError):
            self.call(
                "python -m signac sync {} {} -s never -u".format(
                    os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
                ).split()
            )
        with pytest.raises(ExitCodeError):
            self.call(
                "python -m signac sync {} {} -t".format(
                    os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
                ).split()
            )

    def test_sync_merge(self):
        project_b = signac.init_project(path=os.path.join(self.tmpdir.name, "b"))
        self.call("python -m signac init".split())
        project_a = signac.Project()
        for i in range(4):
            project_a.open_job({"a": i}).init()
            project_b.open_job({"a": i}).init()
        project_a.open_job({"c": 1}).init()
        project_b.open_job({"b": 1}).init()
        project_b.open_job({"a": 4}).init()
        assert len(project_a) == 5
        assert len(project_b) == 6

        # sync with projects having diffent schema
        with pytest.raises(ExitCodeError):
            self.call(
                "python -m signac sync {} {}".format(
                    os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
                ).split()
            )
        assert len(project_a) == 5
        assert len(project_b) == 6

        self.call(
            "python -m signac sync {} {} --merge".format(
                os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
            ).split()
        )
        assert len(project_a) == 7
        assert len(project_b) == 6

    def test_sync_document(self):
        self.call("python -m signac init".split())
        project_a = signac.Project()
        project_b = signac.init_project(path=os.path.join(self.tmpdir.name, "b"))
        job_src = project_a.open_job({"a": 0})
        job_dst = project_b.open_job({"a": 0})

        def reset():
            job_src.document["a"] = 0
            job_src.document["nested"] = dict(a=1)
            job_dst.document["a"] = 1
            job_dst.document["nested"] = dict(a=2)

        # DocumentSyncConflict without any doc-strategy
        reset()
        assert job_dst.document != job_src.document
        with pytest.raises(ExitCodeError):
            self.call(
                "python -m signac sync {} {}".format(
                    os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
                ).split()
            )
        assert job_dst.document != job_src.document
        # don't sync any key
        self.call(
            "python -m signac sync {} {} --no-key".format(
                os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
            ).split()
        )
        assert job_dst.document != job_src.document
        assert job_dst.document["a"] != job_src.document["a"]
        assert job_dst.document["nested"] != job_src.document["nested"]
        # only sync a
        reset()
        self.call(
            "python -m signac sync {} {} --key a".format(
                os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
            ).split()
        )
        assert job_dst.document != job_src.document
        assert job_dst.document["nested"] != job_src.document["nested"]
        assert job_dst.document["a"] == job_src.document["a"]
        # only sync nested
        reset()
        self.call(
            "python -m signac sync {} {} --key nested".format(
                os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
            ).split()
        )
        assert job_dst.document != job_src.document
        assert job_dst.document["a"] != job_src.document["a"]
        assert job_dst.document["nested"] == job_src.document["nested"]
        # sync both
        reset()
        self.call(
            "python -m signac sync {} {} --all-key".format(
                os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
            ).split()
        )
        assert job_dst.document == job_src.document
        assert job_dst.document["nested"] == job_src.document["nested"]
        assert job_dst.document["a"] == job_src.document["a"]
        # invalid input
        with pytest.raises(ExitCodeError):
            self.call(
                "python -m signac sync {} {} --all-key --no-key".format(
                    os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
                ).split()
            )

    def test_sync_file(self):
        self.call("python -m signac init".split())
        project_a = signac.Project()
        project_b = signac.init_project(path=os.path.join(self.tmpdir.name, "b"))
        job_src = project_a.open_job({"a": 0}).init()
        job_dst = project_b.open_job({"a": 0}).init()
        for i, job in enumerate([job_src, job_dst]):
            with open(job.fn("test"), "w") as file:
                file.write("x" * (i + 1))
        # FileSyncConflict
        with pytest.raises(ExitCodeError):
            self.call(
                "python -m signac sync {} {}".format(
                    os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
                ).split()
            )
        self.call(
            "python -m signac sync {} {} --strategy never".format(
                os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
            ).split()
        )
        with open(job_dst.fn("test")) as file:
            assert file.read() == "xx"

        with pytest.raises(ExitCodeError):
            self.call(
                "python -m signac sync {} {}".format(
                    os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
                ).split()
            )
        self.call(
            "python -m signac sync {} {} --update".format(
                os.path.join(self.tmpdir.name, "b"), self.tmpdir.name
            ).split()
        )

    def test_export(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        prefix_data = os.path.join(self.tmpdir.name, "data")

        err = self.call(f"python -m signac export {prefix_data}".split(), error=True)
        assert "No jobs to export" in err

        for i in range(10):
            project.open_job({"a": i}).init()
        assert len(project) == 10
        self.call(f"python -m signac export {prefix_data}".split())
        assert len(project) == 10
        assert len(os.listdir(prefix_data)) == 1
        assert len(os.listdir(os.path.join(prefix_data, "a"))) == 10
        for i in range(10):
            assert os.path.isdir(os.path.join(prefix_data, "a", str(i)))

    def test_import(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        prefix_data = os.path.join(self.tmpdir.name, "data")

        err = self.call(
            f"python -m signac import {self.tmpdir.name}".split(), error=True
        )
        assert "Nothing to import." in err

        for i in range(10):
            project.open_job({"a": i}).init()
        jobs_before_export = {job.id for job in project.find_jobs()}
        assert len(project) == 10
        project.export_to(target=prefix_data, copytree=os.replace)
        assert len(project) == 0
        self.call(f"python -m signac import {prefix_data}".split())
        assert len(project) == 10
        assert {job.id for job in project.find_jobs()} == jobs_before_export

        # invalid combination
        with pytest.raises(ExitCodeError):
            self.call(
                f"python -m signac import {prefix_data} --sync-interactive --move".split()
            )

    def test_import_sync(self):
        project_b = signac.init_project(path=os.path.join(self.tmpdir.name, "b"))
        self.call("python -m signac init".split())
        prefix_data = os.path.join(self.tmpdir.name, "data")
        project_a = signac.Project()
        for i in range(4):
            project_a.open_job({"a": i}).init()
            project_b.open_job({"a": i}).init()
        job_dst = project_a.open_job({"a": 0})
        job_src = project_b.open_job({"a": 0})
        job_src.document["a"] = 0
        project_b.export_to(prefix_data)
        err = self.call(f"python -m signac import {prefix_data}".split(), error=True)
        assert "Import failed" in err
        self.call(f"python -m signac import {prefix_data} --sync".split(), error=True)
        assert len(project_a) == 4
        assert "a" in job_dst.document
        assert job_dst.document["a"] == 0
        out = self.call(
            f"python -m signac import {prefix_data} --sync-interactive",
            "print(str(tmp_project), len(tmp_project)); exit()",
            shell=True,
        )
        assert "4" in out

    def test_shell(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        out = self.call(
            "python -m signac shell",
            "print(str(project), job, len(list(jobs))); exit()",
            shell=True,
        )
        assert out.strip() == f">>> {project} None {len(project)}"

        cmd = "python -m signac shell -c".split() + [
            "print(str(project), len(list(jobs)))"
        ]
        out = self.call(cmd)
        assert out.strip() == f"{project} {len(project)}"

    def test_shell_with_jobs(self):
        out = self.call("python -m signac shell", shell=True)
        assert "No project within this directory" in out

        self.call("python -m signac init".split())
        project = signac.Project()
        for i in range(3):
            project.open_job(dict(a=i)).init()
        assert len(project)
        out = self.call(
            "python -m signac shell",
            "print(str(project), job, len(list(jobs))); exit()",
            shell=True,
        )
        assert out.strip() == f">>> {project} None {len(project)}"

    def test_shell_with_jobs_and_selection(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        for i in range(3):
            project.open_job(dict(a=i)).init()
        assert len(project)
        python_command = "python -m signac shell -f a.{}gt 0".format(
            "$" if WINDOWS else r"\$"
        )
        out = self.call(
            python_command,
            "print(str(project), job, len(list(jobs))); exit()",
            shell=True,
        )
        n = len(project.find_jobs({"a": {"$gt": 0}}))
        assert out.strip() == f">>> {project} None {n}"

    def test_shell_with_jobs_and_selection_only_one_job(self):
        self.call("python -m signac init".split())
        project = signac.Project()
        for i in range(3):
            project.open_job(dict(a=i)).init()
        assert len(project) == 3
        out = self.call(
            "python -m signac shell -f a 0",
            "print(str(project), job, len(list(jobs))); exit()",
            shell=True,
        )
        job = list(project.find_jobs({"a": 0}))[0]
        assert out.strip() == f">>> {project} {job} 1"

    def test_config_show(self):
        err = self.call(
            "python -m signac config --local show".split(), error=True
        ).strip()
        assert "Did not find a local configuration file" in err

        self.call("python -m signac init".split())
        out = self.call("python -m signac config --local show".split()).strip()
        cfg = _read_config_file(".signac/config")
        expected = _Config(cfg).write()
        assert out.split(os.linesep) == expected

        out = self.call("python -m signac config show".split()).strip()
        cfg = _load_config()
        expected = _Config(cfg).write()
        assert out.split(os.linesep) == expected

        out = self.call("python -m signac config --global show".split()).strip()
        cfg = _read_config_file(USER_CONFIG_FN)
        expected = _Config(cfg).write()
        assert out.split(os.linesep) == expected

    def test_config_set(self):
        self.call("python -m signac init".split())
        self.call("python -m signac config set a b".split())
        cfg = self.call("python -m signac config --local show".split())
        assert "a" in cfg
        assert "a = b" in cfg

        self.call("python -m signac config --local set x.y z".split())
        cfg = self.call("python -m signac config --local show".split())
        assert "[x]" in cfg
        assert "y = z" in cfg

        backup_config = os.path.exists(USER_CONFIG_FN)
        global_config_path_backup = USER_CONFIG_FN + ".tmp"
        try:
            # Make a backup of the global config if it exists
            if backup_config:
                shutil.copy2(USER_CONFIG_FN, global_config_path_backup)

            # Test the global config CLI
            self.call("python -m signac config --global set b c".split())
            cfg = self.call("python -m signac config --global show".split())
            assert "b" in cfg
            assert "b = c" in cfg.split(os.linesep)
        finally:
            # Revert the global config to its previous state (or remove it if
            # it did not exist)
            if backup_config:
                shutil.move(global_config_path_backup, USER_CONFIG_FN)
            else:
                os.remove(USER_CONFIG_FN)

    def test_config_verify(self):
        # no config file
        err = self.call("python -m signac config --local verify".split(), error=True)
        assert "Did not find a local configuration file" in err

        self.call("python -m signac init".split())
        err = self.call("python -m signac config --local verify".split(), error=True)
        assert "Passed" in err

    def test_update_cache(self):
        self.call("python -m signac init".split())
        project_a = signac.Project()
        assert not os.path.isfile(project_a.FN_CACHE)

        for i in range(4):
            project_a.open_job({"a": i}).init()
        err = self.call("python -m signac update-cache".split(), error=True)
        assert os.path.isfile(project_a.FN_CACHE)
        assert "Updated cache" in err

        err = self.call("python -m signac update-cache".split(), error=True)
        assert "Cache is up to date" in err

    def test_migrate_v1_to_v2(self):
        dirname = self.tmpdir.name
        _initialize_v1_project(dirname, False)
        self.call("python -m signac migrate --yes".split())
        assert not os.path.isfile(os.path.join(dirname, "signac.rc"))
        assert not os.path.isdir(os.path.join(dirname, "workspace_dir"))
        assert os.path.isdir(os.path.join(dirname, ".signac"))
        assert os.path.isdir(os.path.join(dirname, "workspace"))



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""The signac framework aids in the management of large and heterogeneous data spaces.

It provides a simple and robust data model to create a well-defined, indexable
storage layout for data and metadata. This makes it easier to operate on large
data spaces, streamlines post-processing and analysis, and makes data
collectively accessible.
"""

from . import errors, sync
from ._synced_collections.backends.collection_json import (
    BufferedJSONAttrDict as JSONDict,
)
from .diff import diff_jobs
from .h5store import H5Store, H5StoreManager
from .project import Project, TemporaryProject, get_job, get_project, init_project
from .version import __version__

# Alias some properties related to buffering into the signac namespace.
buffered = JSONDict.buffer_backend
is_buffered = JSONDict.backend_is_buffered
get_current_buffer_size = JSONDict.get_current_buffer_size
get_buffer_capacity = JSONDict.get_buffer_capacity
set_buffer_capacity = JSONDict.set_buffer_capacity

__all__ = [
    "__version__",
    "errors",
    "sync",
    "Project",
    "TemporaryProject",
    "get_project",
    "init_project",
    "get_job",
    "diff_jobs",
    "buffered",
    "is_buffered",
    "get_buffer_capacity",
    "get_current_buffer_size",
    "set_buffer_capacity",
    "JSONDict",
    "H5Store",
    "H5StoreManager",
]
